import datetime
import logging
import os
import time
from decimal import Decimal
from math import ceil, floor, isnan
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd

from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.connector.exchange_base cimport ExchangeBase
from hummingbot.core.clock cimport Clock

from hummingbot.client.config.config_helpers import ClientConfigAdapter
from hummingbot.core.data_type.common import (
    OrderType,
    PriceType,
    TradeType
)
from hummingbot.core.data_type.limit_order cimport LimitOrder
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils import map_df_to_str
from hummingbot.strategy.__utils__.trailing_indicators.instant_volatility import InstantVolatilityIndicator
from hummingbot.strategy.__utils__.trailing_indicators.trading_intensity import TradingIntensityIndicator
from hummingbot.strategy.avellaneda_market_making.avellaneda_market_making_config_map_pydantic import (
    AvellanedaMarketMakingConfigMap,
    DailyBetweenTimesModel,
    FromDateToDateModel,
    MultiOrderLevelModel,
    TrackHangingOrdersModel,
)
from hummingbot.strategy.conditional_execution_state import (
    RunAlwaysExecutionState,
    RunInTimeConditionalExecutionState
)
from hummingbot.strategy.data_types import (
    PriceSize,
    Proposal,
)
from hummingbot.strategy.hanging_orders_tracker import (
    CreatedPairOfOrders,
    HangingOrdersTracker,
)
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.order_book_asset_price_delegate import OrderBookAssetPriceDelegate
from hummingbot.strategy.order_tracker cimport OrderTracker
from hummingbot.strategy.strategy_base import StrategyBase
from hummingbot.strategy.utils import order_age

NaN = float("nan")
s_decimal_zero = Decimal(0)
s_decimal_neg_one = Decimal(-1)
s_decimal_one = Decimal(1)
pmm_logger = None


cdef class AvellanedaMarketMakingStrategy(StrategyBase):
    OPTION_LOG_CREATE_ORDER = 1 << 3
    OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
    OPTION_LOG_STATUS_REPORT = 1 << 5
    OPTION_LOG_ALL = 0x7fffffffffffffff

    @classmethod
    def logger(cls):
        global pmm_logger
        if pmm_logger is None:
            pmm_logger = logging.getLogger(__name__)
        return pmm_logger

    def init_params(self,
                    config_map: Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter],
                    market_info: MarketTradingPairTuple,
                    logging_options: int = OPTION_LOG_ALL,
                    status_report_interval: float = 900,
                    hb_app_notification: bool = False,
                    debug_csv_path: str = '',
                    is_debug: bool = False,
                    ):
        self._sb_order_tracker = OrderTracker()
        self._config_map = config_map
        self._market_info = market_info
        self._price_delegate = OrderBookAssetPriceDelegate(market_info.market, market_info.trading_pair)
        self._hb_app_notification = hb_app_notification
        self._hanging_orders_enabled = False
        self._hanging_orders_cancel_pct = Decimal("10")
        self._hanging_orders_tracker = HangingOrdersTracker(self,
                                                            self._hanging_orders_cancel_pct / Decimal('100'))

        self._cancel_timestamp = 0
        self._create_timestamp = 0
        self._limit_order_type = self._market_info.market.get_maker_order_type()
        self._all_markets_ready = False
        self._filled_buys_balance = 0
        self._filled_sells_balance = 0
        self._logging_options = logging_options
        self._last_timestamp = 0
        self._status_report_interval = status_report_interval
        self._last_own_trade_price = Decimal('nan')

        self.c_add_markets([market_info.market])
        self._volatility_buffer_size = 0
        self._trading_intensity_buffer_size = 0
        self._ticks_to_be_ready = -1
        self._avg_vol = None
        self._trading_intensity = None
        self._last_sampling_timestamp = 0
        self._alpha = None
        self._kappa = None
        self._execution_mode = None
        self._execution_timeframe = None
        self._execution_state = None
        self._start_time = None
        self._end_time = None
        self._reservation_price = s_decimal_zero
        self._optimal_spread = s_decimal_zero
        self._optimal_ask = s_decimal_zero
        self._optimal_bid = s_decimal_zero
        self._debug_csv_path = debug_csv_path
        self._is_debug = is_debug
        try:
            if self._is_debug:
                os.unlink(self._debug_csv_path)
        except FileNotFoundError:
            pass

        self.get_config_map_execution_mode()
        self.get_config_map_hanging_orders()

    def all_markets_ready(self):
        return all([market.ready for market in self._sb_markets])

    @property
    def min_spread(self):
        return self._config_map.min_spread

    @property
    def avg_vol(self):
        return self._avg_vol

    @avg_vol.setter
    def avg_vol(self, indicator: InstantVolatilityIndicator):
        self._avg_vol = indicator

    @property
    def trading_intensity(self):
        return self._trading_intensity

    @trading_intensity.setter
    def trading_intensity(self, indicator: TradingIntensityIndicator):
        self._trading_intensity = indicator

    @property
    def market_info(self) -> MarketTradingPairTuple:
        return self._market_info

    @property
    def order_refresh_tolerance_pct(self) -> Decimal:
        return self._config_map.order_refresh_tolerance_pct

    @property
    def order_refresh_tolerance(self) -> Decimal:
        return self._config_map.order_refresh_tolerance_pct / Decimal('100')

    @property
    def order_amount(self) -> Decimal:
        return self._config_map.order_amount

    @property
    def inventory_target_base_pct(self) -> Decimal:
        return self._config_map.inventory_target_base_pct

    @property
    def inventory_target_base(self) -> Decimal:
        return self.inventory_target_base_pct / Decimal('100')

    @inventory_target_base.setter
    def inventory_target_base(self, value: Decimal):
        self.inventory_target_base_pct = value * Decimal('100')

    @property
    def order_optimization_enabled(self) -> bool:
        return self._config_map.order_optimization_enabled

    @property
    def order_refresh_time(self) -> float:
        return self._config_map.order_refresh_time

    @property
    def filled_order_delay(self) -> float:
        return self._config_map.filled_order_delay

    @property
    def order_override(self) -> Dict[str, any]:
        return self._config_map.order_override

    @property
    def order_levels(self) -> int:
        if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:
            return self._config_map.order_levels_mode.order_levels
        else:
            return 0

    @property
    def level_distances(self) -> int:
        if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:
            return self._config_map.order_levels_mode.level_distances
        else:
            return 0

    @property
    def max_order_age(self):
        return self._config_map.max_order_age

    @property
    def add_transaction_costs_to_orders(self) -> bool:
        return self._config_map.add_transaction_costs

    @property
    def base_asset(self):
        return self._market_info.base_asset

    @property
    def quote_asset(self):
        return self._market_info.quote_asset

    @property
    def trading_pair(self):
        return self._market_info.trading_pair

    @property
    def gamma(self):
        return self._config_map.risk_factor

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        self._alpha = value

    @property
    def kappa(self):
        return self._kappa

    @kappa.setter
    def kappa(self, value):
        self._kappa = value

    @property
    def eta(self):
        return self._config_map.order_amount_shape_factor

    @property
    def reservation_price(self):
        return self._reservation_price

    @reservation_price.setter
    def reservation_price(self, value):
        self._reservation_price = value

    @property
    def optimal_spread(self):
        return self._optimal_spread

    @property
    def optimal_ask(self):
        return self._optimal_ask

    @optimal_ask.setter
    def optimal_ask(self, value):
        self._optimal_ask = value

    @property
    def optimal_bid(self):
        return self._optimal_bid

    @optimal_bid.setter
    def optimal_bid(self, value):
        self._optimal_bid = value

    @property
    def execution_timeframe(self):
        return self._execution_timeframe

    @property
    def start_time(self) -> time:
        return self._start_time

    @start_time.setter
    def start_time(self, value):
        self._start_time = value

    @property
    def end_time(self) -> time:
        return self._end_time

    @end_time.setter
    def end_time(self, value):
        self._end_time = value

    def get_price(self) -> float:
        return self.get_mid_price()

    def get_last_price(self) -> float:
        return self._market_info.get_last_price()

    def get_mid_price(self) -> float:
        return self.c_get_mid_price()

    cdef object c_get_mid_price(self):
        return self._price_delegate.get_price_by_type(PriceType.MidPrice)

    @property
    def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
        return self._sb_order_tracker.market_pair_to_active_orders

    @property
    def active_orders(self) -> List[LimitOrder]:
        if self._market_info not in self.market_info_to_active_orders:
            return []
        return self.market_info_to_active_orders[self._market_info]

    @property
    def active_non_hanging_orders(self) -> List[LimitOrder]:
        orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]
        return orders

    @property
    def active_buys(self) -> List[LimitOrder]:
        return [o for o in self.active_orders if o.is_buy]

    @property
    def active_sells(self) -> List[LimitOrder]:
        return [o for o in self.active_orders if not o.is_buy]

    @property
    def logging_options(self) -> int:
        return self._logging_options

    @logging_options.setter
    def logging_options(self, int64_t logging_options):
        self._logging_options = logging_options

    @property
    def hanging_orders_tracker(self):
        return self._hanging_orders_tracker

    def update_from_config_map(self):
        self.get_config_map_execution_mode()
        self.get_config_map_hanging_orders()
        self.get_config_map_indicators()

    def get_config_map_execution_mode(self):
        try:
            execution_mode = self._config_map.execution_timeframe_mode.title
            execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]
            if execution_mode == FromDateToDateModel.model_config["title"]:
                start_time = self._config_map.execution_timeframe_mode.start_datetime
                end_time = self._config_map.execution_timeframe_mode.end_datetime
                execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
            elif execution_mode == DailyBetweenTimesModel.model_config["title"]:
                start_time = self._config_map.execution_timeframe_mode.start_time
                end_time = self._config_map.execution_timeframe_mode.end_time
                execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
            else:
                start_time = None
                end_time = None
                execution_state = RunAlwaysExecutionState()

            # Something has changed?
            if self._execution_state is None or self._execution_state != execution_state:
                self._execution_state = execution_state
                self._execution_mode = execution_mode
                self._execution_timeframe = execution_timeframe
                self._start_time = start_time
                self._end_time = end_time
        except AttributeError:
            # A parameter is missing in the execution timeframe mode configuration
            pass

    def get_config_map_hanging_orders(self):
        if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:
            hanging_orders_enabled = True
            hanging_orders_cancel_pct = self._config_map.hanging_orders_mode.hanging_orders_cancel_pct
        else:
            hanging_orders_enabled = False
            hanging_orders_cancel_pct = Decimal("0")

        if self._hanging_orders_enabled != hanging_orders_enabled:
            # Hanging order tracker instance doesn't exist - create from scratch
            self._hanging_orders_enabled = hanging_orders_enabled
            self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
            self._hanging_orders_tracker = HangingOrdersTracker(self,
                                                                hanging_orders_cancel_pct / Decimal('100'))
            self._hanging_orders_tracker.register_events(self.active_markets)
        elif self._hanging_orders_cancel_pct != hanging_orders_cancel_pct:
            # Hanging order tracker instance existst - only update variable
            self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
            self._hanging_orders_tracker.hanging_orders_cancel_pct = hanging_orders_cancel_pct / Decimal('100')

    def get_config_map_indicators(self):
        volatility_buffer_size = self._config_map.volatility_buffer_size
        trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size
        ticks_to_be_ready_after = max(volatility_buffer_size, trading_intensity_buffer_size)
        ticks_to_be_ready_before = max(self._volatility_buffer_size, self._trading_intensity_buffer_size)

        if self._volatility_buffer_size == 0 or self._volatility_buffer_size != volatility_buffer_size:
            self._volatility_buffer_size = volatility_buffer_size

            if self._avg_vol is None:
                self._avg_vol = InstantVolatilityIndicator(sampling_length=volatility_buffer_size)
            else:
                self._avg_vol.sampling_length = volatility_buffer_size

        if (
            self._trading_intensity_buffer_size == 0
            or self._trading_intensity_buffer_size != trading_intensity_buffer_size
        ):
            self._trading_intensity_buffer_size = trading_intensity_buffer_size
            if self._trading_intensity is not None:
                self._trading_intensity.sampling_length = trading_intensity_buffer_size

        if self._trading_intensity is None and self.market_info.market.ready:
            self._trading_intensity = TradingIntensityIndicator(
                order_book=self.market_info.order_book,
                price_delegate=self._price_delegate,
                sampling_length=self._trading_intensity_buffer_size,
            )

        self._ticks_to_be_ready += (ticks_to_be_ready_after - ticks_to_be_ready_before)
        if self._ticks_to_be_ready < 0:
            self._ticks_to_be_ready = 0

    def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:
        market, trading_pair, base_asset, quote_asset = self._market_info
        price = self._price_delegate.get_price_by_type(PriceType.MidPrice)
        base_balance = float(market.get_balance(base_asset))
        quote_balance = float(market.get_balance(quote_asset))
        available_base_balance = float(market.get_available_balance(base_asset))
        available_quote_balance = float(market.get_available_balance(quote_asset))
        base_value = base_balance * float(price)
        total_in_quote = base_value + quote_balance
        base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
        quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
        data = [
            ["", base_asset, quote_asset],
            ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
            ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],
            [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
        ]
        if to_show_current_pct:
            data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
        df = pd.DataFrame(data=data)
        return df

    def active_orders_df(self) -> pd.DataFrame:
        market, trading_pair, base_asset, quote_asset = self._market_info
        price = self.get_price()
        active_orders = self.active_orders
        no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
                        not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
        active_orders.sort(key=lambda x: x.price, reverse=True)
        columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
        data = []
        lvl_buy, lvl_sell = 0, 0
        for idx in range(0, len(active_orders)):
            order = active_orders[idx]
            is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
            if not is_hanging_order:
                if order.is_buy:
                    level = lvl_buy + 1
                    lvl_buy += 1
                else:
                    level = no_sells - lvl_sell
                    lvl_sell += 1
            spread = 0 if price == 0 else abs(order.price - price) / price
            age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')

            amount_orig = self._config_map.order_amount
            if is_hanging_order:
                amount_orig = order.quantity
                level = "hang"
            data.append([
                level,
                "buy" if order.is_buy else "sell",
                float(order.price),
                f"{spread:.2%}",
                float(amount_orig),
                float(order.quantity),
                age
            ])

        return pd.DataFrame(data=data, columns=columns)

    def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
        markets_data = []
        markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]
        markets_columns.append('Reservation Price')
        markets_columns.append('Optimal Spread')
        market_books = [(self._market_info.market, self._market_info.trading_pair)]
        for market, trading_pair in market_books:
            bid_price = market.get_price(trading_pair, False)
            ask_price = market.get_price(trading_pair, True)
            ref_price = self.get_price()
            markets_data.append([
                market.display_name,
                trading_pair,
                float(bid_price),
                float(ask_price),
                float(ref_price),
                round(self._reservation_price, 5),
                round(self._optimal_spread, 5),
            ])
        return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)

    def format_status(self) -> str:
        if not self._all_markets_ready:
            return "Market connectors are not ready."
        cdef:
            list lines = []
            list warning_lines = []
        warning_lines.extend(self.network_warning([self._market_info]))

        markets_df = self.market_status_data_frame([self._market_info])
        lines.extend(["", "  Markets:"] + ["    " + line for line in markets_df.to_string(index=False).split("\n")])

        assets_df = map_df_to_str(self.pure_mm_assets_df(True))
        first_col_length = max(*assets_df[0].apply(len))
        df_lines = assets_df.to_string(index=False, header=False,
                                       formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")
        lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])

        # See if there are any open orders.
        if len(self.active_orders) > 0:
            df = self.active_orders_df()
            lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
        else:
            lines.extend(["", "  No active maker orders."])

        volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
        if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):
            lines.extend(["", f"  Strategy parameters:",
                          f"    risk_factor(\u03B3)= {self.gamma:.5E}",
                          f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
                          f"    order_book_depth_factor(\u03BA)= {self._kappa:.5E}",
                          f"    volatility= {volatility_pct:.3f}%"])
            if self._execution_state.time_left is not None:
                lines.extend([f"    time until end of trading cycle = {str(datetime.timedelta(seconds=float(self._execution_state.time_left)//1e3))}"])
            else:
                lines.extend([f"    time until end of trading cycle = N/A"])

        warning_lines.extend(self.balance_warning([self._market_info]))

        if len(warning_lines) > 0:
            lines.extend(["", "*** WARNINGS ***"] + warning_lines)

        return "\n".join(lines)

    def execute_orders_proposal(self, proposal: Proposal):
        return self.c_execute_orders_proposal(proposal)

    def cancel_order(self, order_id: str):
        return self.c_cancel_order(self._market_info, order_id)

    cdef c_start(self, Clock clock, double timestamp):
        StrategyBase.c_start(self, clock, timestamp)
        self.update_from_config_map()
        self._last_timestamp = timestamp

        self._hanging_orders_tracker.register_events(self.active_markets)

        if self._hanging_orders_enabled:
            # start tracking any restored limit order
            restored_order_ids = self.c_track_restored_orders(self.market_info)
            for order_id in restored_order_ids:
                order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
                if order:
                    self._hanging_orders_tracker.add_as_hanging_order(order)

        self._execution_state.time_left = self._execution_state.closing_time

    def start(self, clock: Clock, timestamp: float):
        self.c_start(clock, timestamp)

    cdef c_stop(self, Clock clock):
        self._hanging_orders_tracker.unregister_events(self.active_markets)
        StrategyBase.c_stop(self, clock)

    cdef c_tick(self, double timestamp):
        StrategyBase.c_tick(self, timestamp)
        cdef:
            int64_t current_tick = <int64_t>(timestamp // self._status_report_interval)
            int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)
            bint should_report_warnings = ((current_tick > last_tick) and
                                           (self._logging_options & self.OPTION_LOG_STATUS_REPORT))
            object proposal

        try:
            if not self._all_markets_ready:
                self._all_markets_ready = all([mkt.ready for mkt in self._sb_markets])
                if not self._all_markets_ready:
                    # Markets not ready yet. Don't do anything.
                    if should_report_warnings:
                        self.logger().warning(f"Markets are not ready. No market making trades are permitted.")
                    return

            if should_report_warnings:
                if not all([mkt.network_status is NetworkStatus.CONNECTED for mkt in self._sb_markets]):
                    self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
                                          f"making may be dangerous when markets or networks are unstable.")

            # Updates settings from config map if changed
            self.update_from_config_map()

            self.c_collect_market_variables(timestamp)

            if self.c_is_algorithm_ready():
                if self._create_timestamp <= self._current_timestamp:
                    # Measure order book liquidity
                    self.c_measure_order_book_liquidity()

                self._hanging_orders_tracker.process_tick()

                # Needs to be executed at all times to not to have active order leftovers after a trading session ends
                self.c_cancel_active_orders_on_max_age_limit()

                # process_tick() is only called if within a trading timeframe
                self._execution_state.process_tick(timestamp, self)

            else:
                # Only if snapshots are different - for trading intensity - a market order happened
                if self.c_is_algorithm_changed():
                    self._ticks_to_be_ready -= 1
                    if self._ticks_to_be_ready % 5 == 0:
                        self.logger().info(f"Calculating volatility, estimating order book liquidity ... {self._ticks_to_be_ready} ticks to fill buffers")
                else:
                    self.logger().info(f"Calculating volatility, estimating order book liquidity ... no trades tick")
        finally:
            self._last_timestamp = timestamp

    def process_tick(self, timestamp: float):
        proposal = None
        # Trading is allowed
        if self._create_timestamp <= self._current_timestamp:
            # 1. Calculate reservation price and optimal spread from gamma, alpha, kappa and volatility
            self.c_calculate_reservation_price_and_optimal_spread()
            # 2. Check if calculated prices make sense
            if self._optimal_bid > 0 and self._optimal_ask > 0:
                # 3. Create base order proposals
                proposal = self.c_create_base_proposal()
                # 4. Apply functions that modify orders amount
                self.c_apply_order_amount_eta_transformation(proposal)
                # 5. Apply functions that modify orders price
                self.c_apply_order_price_modifiers(proposal)
                # 6. Apply budget constraint, i.e. can't buy/sell more than what you have.
                self.c_apply_budget_constraint(proposal)

                self.c_cancel_active_orders(proposal)

        if self.c_to_create_orders(proposal):
            self.c_execute_orders_proposal(proposal)

        if self._is_debug:
            self.dump_debug_variables()

    cdef c_collect_market_variables(self, double timestamp):
        market, trading_pair, base_asset, quote_asset = self._market_info
        self._last_sampling_timestamp = timestamp

        price = self.get_price()
        self._avg_vol.add_sample(price)
        self._trading_intensity.calculate(timestamp)
        # Calculate adjustment factor to have 0.01% of inventory resolution
        base_balance = market.get_balance(base_asset)
        quote_balance = market.get_balance(quote_asset)
        inventory_in_base = quote_balance / price + base_balance

    def collect_market_variables(self, timestamp: float):
        self.c_collect_market_variables(timestamp)

    cdef double c_get_spread(self):
        cdef:
            ExchangeBase market = self._market_info.market
            str trading_pair = self._market_info.trading_pair

        return market.c_get_price(trading_pair, True) - market.c_get_price(trading_pair, False)

    def get_spread(self):
        return self.c_get_spread()

    def get_volatility(self):
        vol = Decimal(str(self._avg_vol.current_value))
        return vol

    cdef c_measure_order_book_liquidity(self):

        self._alpha, self._kappa = self._trading_intensity.current_value

        self._alpha = Decimal(self._alpha)
        self._kappa = Decimal(self._kappa)

        if self._is_debug:
            self.logger().info(f"alpha={self._alpha:.4f} | "
                               f"kappa={self._kappa:.4f}")

    def measure_order_book_liquidity(self):
        return self.c_measure_order_book_liquidity()

    cdef c_calculate_reservation_price_and_optimal_spread(self):
        cdef:
            ExchangeBase market = self._market_info.market

        # Current mid price
        price = self.get_price()

        # The amount of stocks owned - q - has to be in relative units, not absolute, because changing the portfolio size shouldn't change the reservation price
        # The reservation price should concern itself only with the strategy performance, i.e. amount of stocks relative to the target
        inventory = Decimal(str(self.c_calculate_inventory()))
        if inventory == 0:
            return

        q_target = Decimal(str(self.c_calculate_target_inventory()))
        q = (market.get_balance(self.base_asset) - q_target) / (inventory)
        # Volatility has to be in absolute values (prices) because in calculation of reservation price it's not multiplied by the current price, therefore
        # it can't be a percentage. The result of the multiplication has to be an absolute price value because it's being subtracted from the current price
        vol = self.get_volatility()

        # order book liquidity - kappa and alpha have to represent absolute values because the second member of the optimal spread equation has to be an absolute price
        # and from the reservation price calculation we know that gamma's unit is not absolute price
        if all((self.gamma, self._kappa)) and self._alpha != 0 and self._kappa > 0 and vol != 0:
            if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:
                # Avellaneda-Stoikov for a fixed timespan
                time_left_fraction = Decimal(str(self._execution_state.time_left / self._execution_state.closing_time))
            else:
                # Avellaneda-Stoikov for an infinite timespan
                # The equations in the paper for this contain a few mistakes
                # - the units don't align with the rest of the paper
                # - volatility cancels itself out completely
                # - the risk factor gets partially canceled
                # The proposed solution is to use the same equation as for the constrained timespan but with
                # a fixed time left
                time_left_fraction = 1

            # Here seems to be another mistake in the paper
            # It doesn't make sense to use mid_price_variance because its units would be absolute price units ^2, yet that side of the equation is subtracted
            # from the actual mid price of the asset in absolute price units
            # gamma / risk_factor gains a meaning of a fraction (percentage) of the volatility (standard deviation between ticks) to be subtraced from the
            # current mid price
            # This leads to normalization of the risk_factor and will guaranetee consistent behavior on all price ranges of the asset, and across assets

            self._reservation_price = price - (q * self.gamma * vol * time_left_fraction)

            self._optimal_spread = self.gamma * vol * time_left_fraction
            self._optimal_spread += 2 * Decimal(1 + self.gamma / self._kappa).ln() / self.gamma

            min_spread = price / 100 * Decimal(str(self._config_map.min_spread))

            max_limit_bid = price - min_spread / 2
            min_limit_ask = price + min_spread / 2

            self._optimal_ask = max(self._reservation_price + self._optimal_spread / 2, min_limit_ask)
            self._optimal_bid = min(self._reservation_price - self._optimal_spread / 2, max_limit_bid)

            # This is not what the algorithm will use as proposed bid and ask. This is just the raw output.
            # Optimal bid and optimal ask prices will be used
            if self._is_debug:
                self.logger().info(f"q={q:.4f} | "
                                   f"vol={vol:.10f}")
                self.logger().info(f"mid_price={price:.10f} | "
                                   f"reservation_price={self._reservation_price:.10f} | "
                                   f"optimal_spread={self._optimal_spread:.10f}")
                self.logger().info(f"optimal_bid={(price-(self._reservation_price - self._optimal_spread / 2)) / price * 100:.4f}% | "
                                   f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")

    def calculate_reservation_price_and_optimal_spread(self):
        return self.c_calculate_reservation_price_and_optimal_spread()

    cdef object c_calculate_target_inventory(self):
        cdef:
            ExchangeBase market = self._market_info.market
            str trading_pair = self._market_info.trading_pair
            str base_asset = self._market_info.base_asset
            str quote_asset = self._market_info.quote_asset
            object mid_price
            object base_value
            object inventory_value
            object target_inventory_value

        price = self.get_price()
        base_asset_amount = market.get_balance(base_asset)
        quote_asset_amount = market.get_balance(quote_asset)
        # Base asset value in quote asset prices
        base_value = base_asset_amount * price
        # Total inventory value in quote asset prices
        inventory_value = base_value + quote_asset_amount
        # Target base asset value in quote asset prices
        target_inventory_value = inventory_value * self.inventory_target_base
        # Target base asset amount
        target_inventory_amount = target_inventory_value / price
        return market.c_quantize_order_amount(trading_pair, Decimal(str(target_inventory_amount)))

    def calculate_target_inventory(self) -> Decimal:
        return self.c_calculate_target_inventory()

    cdef c_calculate_inventory(self):
        cdef:
            ExchangeBase market = self._market_info.market
            str base_asset = self._market_info.base_asset
            str quote_asset = self._market_info.quote_asset
            object mid_price
            object base_value
            object inventory_value
            object inventory_value_quote
            object inventory_value_base

        price = self.get_price()
        base_asset_amount = market.get_balance(base_asset)
        quote_asset_amount = market.get_balance(quote_asset)
        # Base asset value in quote asset prices
        base_value = base_asset_amount * price
        # Total inventory value in quote asset prices
        inventory_value_quote = base_value + quote_asset_amount
        # Total inventory value in base asset prices
        inventory_value_base = inventory_value_quote / price
        return inventory_value_base

    def calculate_inventory(self) -> Decimal:
        return self.c_calculate_inventory()

    cdef bint c_is_algorithm_ready(self):
        return self._avg_vol.is_sampling_buffer_full and self._trading_intensity.is_sampling_buffer_full

    cdef bint c_is_algorithm_changed(self):
        return self._trading_intensity.is_sampling_buffer_changed or self._avg_vol.is_sampling_buffer_changed

    def is_algorithm_ready(self) -> bool:
        return self.c_is_algorithm_ready()

    def is_algorithm_changed(self) -> bool:
        return self.c_is_algorithm_changed()

    def _get_level_spreads(self):
        level_step = ((self._optimal_spread / 2) / 100) * self.level_distances

        bid_level_spreads = [i * level_step for i in range(self.order_levels)]
        ask_level_spreads = [i * level_step for i in range(self.order_levels)]

        return bid_level_spreads, ask_level_spreads

    cdef _create_proposal_based_on_order_override(self):
        cdef:
            ExchangeBase market = self._market_info.market
            list buys = []
            list sells = []
        reference_price = self.get_price()
        if self.order_override is not None:
            for key, value in self.order_override.items():
                if str(value[0]) in ["buy", "sell"]:
                    list_to_be_appended = buys if str(value[0]) == "buy" else sells
                    size = Decimal(str(value[2]))
                    size = market.c_quantize_order_amount(self.trading_pair, size)
                    if str(value[0]) == "buy":
                        price = reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
                    elif str(value[0]) == "sell":
                        price = reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
                    price = market.c_quantize_order_price(self.trading_pair, price)
                    if size > 0 and price > 0:
                        list_to_be_appended.append(PriceSize(price, size))
        return buys, sells

    def create_proposal_based_on_order_override(self) -> Tuple[List[Proposal], List[Proposal]]:
        return self._create_proposal_based_on_order_override()

    cdef _create_proposal_based_on_order_levels(self):
        cdef:
            ExchangeBase market = self._market_info.market
            list buys = []
            list sells = []
        bid_level_spreads, ask_level_spreads = self._get_level_spreads()
        size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
        if size > 0:
            for level in range(self.order_levels):
                bid_price = market.c_quantize_order_price(self.trading_pair,
                                                          self._optimal_bid - Decimal(str(bid_level_spreads[level])))
                ask_price = market.c_quantize_order_price(self.trading_pair,
                                                          self._optimal_ask + Decimal(str(ask_level_spreads[level])))

                buys.append(PriceSize(bid_price, size))
                sells.append(PriceSize(ask_price, size))
        return buys, sells

    def create_proposal_based_on_order_levels(self):
        return self._create_proposal_based_on_order_levels()

    cdef _create_basic_proposal(self):
        cdef:
            ExchangeBase market = self._market_info.market
            list buys = []
            list sells = []
        price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_bid)))
        size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
        if size > 0:
            buys.append(PriceSize(price, size))

        price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_ask)))
        size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
        if size > 0:
            sells.append(PriceSize(price, size))
        return buys, sells

    def create_basic_proposal(self):
        return self._create_basic_proposal()

    cdef object c_create_base_proposal(self):
        cdef:
            ExchangeBase market = self._market_info.market
            list buys = []
            list sells = []

        if self.order_override is not None and len(self.order_override) > 0:
            # If order_override is set, it will override order_levels
            buys, sells = self._create_proposal_based_on_order_override()
        elif self.order_levels > 0:
            # Simple order levels
            buys, sells = self._create_proposal_based_on_order_levels()
        else:
            # No order levels nor order_overrides. Just 1 bid and 1 ask order
            buys, sells = self._create_basic_proposal()

        return Proposal(buys, sells)

    def create_base_proposal(self):
        return self.c_create_base_proposal()

    cdef tuple c_get_adjusted_available_balance(self, list orders):
        """
        Calculates the available balance, plus the amount attributed to orders.
        :return: (base amount, quote amount) in Decimal
        """
        cdef:
            ExchangeBase market = self._market_info.market
            object base_balance = market.c_get_available_balance(self.base_asset)
            object quote_balance = market.c_get_available_balance(self.quote_asset)

        for order in orders:
            if order.is_buy:
                quote_balance += order.quantity * order.price
            else:
                base_balance += order.quantity

        return base_balance, quote_balance

    def get_adjusted_available_balance(self, orders: List[LimitOrder]):
        return self.c_get_adjusted_available_balance(orders)

    cdef c_apply_order_price_modifiers(self, object proposal):
        if self._config_map.order_optimization_enabled:
            self.c_apply_order_optimization(proposal)

        if self._config_map.add_transaction_costs:
            self.c_apply_add_transaction_costs(proposal)

    def apply_order_price_modifiers(self, proposal: Proposal):
        self.c_apply_order_price_modifiers(proposal)

    def apply_budget_constraint(self, proposal: Proposal):
        return self.c_apply_budget_constraint(proposal)

    def adjusted_available_balance_for_orders_budget_constrain(self):
        candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
        non_hanging = []
        if self.market_info in self._sb_order_tracker.get_limit_orders():
            all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
            non_hanging = [order for order in all_orders
                           if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
        all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
        return self.c_get_adjusted_available_balance(all_non_hanging_orders)

    cdef c_apply_budget_constraint(self, object proposal):
        cdef:
            ExchangeBase market = self._market_info.market
            object quote_size
            object base_size
            object adjusted_amount

        base_balance, quote_balance = self.adjusted_available_balance_for_orders_budget_constrain()

        for buy in proposal.buys:
            buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
                                       buy.size, buy.price)
            quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)

            # Adjust buy order size to use remaining balance if less than the order amount
            if quote_balance < quote_size:
                adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
                adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
                buy.size = adjusted_amount
                quote_balance = s_decimal_zero
            elif quote_balance == s_decimal_zero:
                buy.size = s_decimal_zero
            else:
                quote_balance -= quote_size

        proposal.buys = [o for o in proposal.buys if o.size > 0]

        for sell in proposal.sells:
            base_size = sell.size

            # Adjust sell order size to use remaining balance if less than the order amount
            if base_balance < base_size:
                adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
                sell.size = adjusted_amount
                base_balance = s_decimal_zero
            elif base_balance == s_decimal_zero:
                sell.size = s_decimal_zero
            else:
                base_balance -= base_size

        proposal.sells = [o for o in proposal.sells if o.size > 0]

    def apply_budget_constraint(self, proposal: Proposal):
        return self.c_apply_budget_constraint(proposal)

    # Compare the market price with the top bid and top ask price
    cdef c_apply_order_optimization(self, object proposal):
        cdef:
            ExchangeBase market = self._market_info.market
            object own_buy_size = s_decimal_zero
            object own_sell_size = s_decimal_zero
            object best_order_spread

        for order in self.active_orders:
            if order.is_buy:
                own_buy_size = order.quantity
            else:
                own_sell_size = order.quantity

        if len(proposal.buys) > 0:
            # Get the top bid price in the market using order_optimization_depth and your buy order volume
            top_bid_price = self._market_info.get_price_for_volume(
                False, own_buy_size).result_price
            price_quantum = market.c_get_order_price_quantum(
                self.trading_pair,
                top_bid_price
            )
            # Get the price above the top bid
            price_above_bid = (ceil(top_bid_price / price_quantum) + 1) * price_quantum

            # If the price_above_bid is lower than the price suggested by the top pricing proposal,
            # lower the price and from there apply the best_order_spread to each order in the next levels
            proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)
            for i, proposed in enumerate(proposal.buys):
                if proposal.buys[i].price > price_above_bid:
                    proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, price_above_bid)

        if len(proposal.sells) > 0:
            # Get the top ask price in the market using order_optimization_depth and your sell order volume
            top_ask_price = self._market_info.get_price_for_volume(
                True, own_sell_size).result_price
            price_quantum = market.c_get_order_price_quantum(
                self.trading_pair,
                top_ask_price
            )
            # Get the price below the top ask
            price_below_ask = (floor(top_ask_price / price_quantum) - 1) * price_quantum

            # If the price_below_ask is higher than the price suggested by the pricing proposal,
            # increase your price and from there apply the best_order_spread to each order in the next levels
            proposal.sells = sorted(proposal.sells, key = lambda p: p.price)
            for i, proposed in enumerate(proposal.sells):
                if proposal.sells[i].price < price_below_ask:
                    proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)

    def apply_order_optimization(self, proposal: Proposal):
        return self.c_apply_order_optimization(proposal)

    cdef c_apply_order_amount_eta_transformation(self, object proposal):
        cdef:
            ExchangeBase market = self._market_info.market
            str trading_pair = self._market_info.trading_pair

        # Order amounts should be changed only if order_override is not active
        if (self.order_override is None) or (len(self.order_override) == 0):
            # eta parameter is described in the paper as the shape parameter for having exponentially decreasing order amount
            # for orders that go against inventory target (i.e. Want to buy when excess inventory or sell when deficit inventory)

            # q cannot be in absolute values - cannot be dependent on the size of the inventory or amount of base asset
            # because it's a scaling factor
            inventory = Decimal(str(self.c_calculate_inventory()))

            if inventory == 0:
                return

            q_target = Decimal(str(self.c_calculate_target_inventory()))
            q = (market.get_balance(self.base_asset) - q_target) / (inventory)

            if len(proposal.buys) > 0:
                if q > 0:
                    for i, proposed in enumerate(proposal.buys):

                        proposal.buys[i].size = market.c_quantize_order_amount(trading_pair, proposal.buys[i].size * Decimal.exp(-self.eta * q))
                    proposal.buys = [o for o in proposal.buys if o.size > 0]

            if len(proposal.sells) > 0:
                if q < 0:
                    for i, proposed in enumerate(proposal.sells):
                        proposal.sells[i].size = market.c_quantize_order_amount(trading_pair, proposal.sells[i].size * Decimal.exp(self.eta * q))
                    proposal.sells = [o for o in proposal.sells if o.size > 0]

    def apply_order_amount_eta_transformation(self, proposal: Proposal):
        self.c_apply_order_amount_eta_transformation(proposal)

    cdef c_apply_add_transaction_costs(self, object proposal):
        cdef:
            ExchangeBase market = self._market_info.market
        for buy in proposal.buys:
            fee = market.c_get_fee(self.base_asset, self.quote_asset,
                                   self._limit_order_type, TradeType.BUY, buy.size, buy.price)
            price = buy.price * (Decimal(1) - fee.percent)
            buy.price = market.c_quantize_order_price(self.trading_pair, price)
        for sell in proposal.sells:
            fee = market.c_get_fee(self.base_asset, self.quote_asset,
                                   self._limit_order_type, TradeType.SELL, sell.size, sell.price)
            price = sell.price * (Decimal(1) + fee.percent)
            sell.price = market.c_quantize_order_price(self.trading_pair, price)

    def apply_add_transaction_costs(self, proposal: Proposal):
        self.c_apply_add_transaction_costs(proposal)

    cdef c_did_fill_order(self, object order_filled_event):
        cdef:
            str order_id = order_filled_event.order_id
            object market_info = self._sb_order_tracker.c_get_shadow_market_pair_from_order_id(order_id)
            tuple order_fill_record

        if market_info is not None:
            limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)
            order_fill_record = (limit_order_record, order_filled_event)

            if order_filled_event.trade_type is TradeType.BUY:
                if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
                    self.log_with_clock(
                        logging.INFO,
                        f"({market_info.trading_pair}) Maker buy order of "
                        f"{order_filled_event.amount} {market_info.base_asset} filled."
                    )
            else:
                if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
                    self.log_with_clock(
                        logging.INFO,
                        f"({market_info.trading_pair}) Maker sell order of "
                        f"{order_filled_event.amount} {market_info.base_asset} filled."
                    )

    cdef c_did_complete_buy_order(self, object order_completed_event):
        self.c_did_complete_order(order_completed_event)

    cdef c_did_complete_sell_order(self, object order_completed_event):
        self.c_did_complete_order(order_completed_event)

    cdef c_did_complete_order(self, object order_completed_event):
        cdef:
            str order_id = order_completed_event.order_id
            LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)

        if limit_order_record is None:
            return

        # Continue only if the order is not a hanging order
        if (not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order_id)
                and not self.hanging_orders_tracker.is_order_id_in_completed_hanging_orders(order_id)):
            # delay order creation by filled_order_delay (in seconds)
            self._create_timestamp = self._current_timestamp + self.filled_order_delay
            self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)

            if limit_order_record.is_buy:
                self._filled_buys_balance += 1
                order_action_string = "buy"
            else:
                self._filled_sells_balance += 1
                order_action_string = "sell"

            self._last_own_trade_price = limit_order_record.price

            self.log_with_clock(
                logging.INFO,
                f"({self.trading_pair}) Maker {order_action_string} order {order_id} "
                f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
                f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
            )
            self.notify_hb_app_with_timestamp(
                f"Maker {order_action_string.upper()} order "
                f"{limit_order_record.quantity} {limit_order_record.base_currency} @ "
                f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
            )

    cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):
        if len(current_prices) != len(proposal_prices):
            return False
        current_prices = sorted(current_prices)
        proposal_prices = sorted(proposal_prices)
        for current, proposal in zip(current_prices, proposal_prices):
            # if spread diff is more than the tolerance or order quantities are different, return false.
            if abs(proposal - current) / current > self.order_refresh_tolerance:
                return False
        return True

    def is_within_tolerance(self, current_prices: List[Decimal], proposal_prices: List[Decimal]) -> bool:
        return self.c_is_within_tolerance(current_prices, proposal_prices)

    cdef c_cancel_active_orders_on_max_age_limit(self):
        """
        Cancels active non hanging orders if they are older than max age limit
        """
        cdef:
            list active_orders = self.active_non_hanging_orders
        for order in active_orders:
            if order_age(order, self._current_timestamp) > self._config_map.max_order_age:
                self.c_cancel_order(self._market_info, order.client_order_id)

    cdef c_cancel_active_orders(self, object proposal):
        if self._cancel_timestamp > self._current_timestamp:
            return

        cdef:
            list active_buy_prices = []
            list active_sells = []
            bint to_defer_canceling = False

        if len(self.active_non_hanging_orders) == 0:
            return
        if proposal is not None:
            active_buy_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if o.is_buy]
            active_sell_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if not o.is_buy]
            proposal_buys = [buy.price for buy in proposal.buys]
            proposal_sells = [sell.price for sell in proposal.sells]

            if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \
                    self.c_is_within_tolerance(active_sell_prices, proposal_sells):
                to_defer_canceling = True

        if not to_defer_canceling:
            self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
            for order in self.active_non_hanging_orders:
                # If is about to be added to hanging_orders then don't cancel
                if not self._hanging_orders_tracker.is_potential_hanging_order(order):
                    self.c_cancel_order(self._market_info, order.client_order_id)
        else:
            self.c_set_timers()

    def cancel_active_orders(self, proposal: Proposal = None):
        return self.c_cancel_active_orders(proposal)

    cdef bint c_to_create_orders(self, object proposal):
        non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not
                                            self._hanging_orders_tracker.is_potential_hanging_order(o)]

        return (self._create_timestamp < self._current_timestamp
                and (not self._config_map.should_wait_order_cancel_confirmation or
                     len(self._sb_order_tracker.in_flight_cancels) == 0)
                and proposal is not None
                and len(non_hanging_orders_non_cancelled) == 0)

    def to_create_orders(self, proposal: Proposal) -> bool:
        return self.c_to_create_orders(proposal)

    cdef c_execute_orders_proposal(self, object proposal):
        cdef:
            double expiration_seconds = NaN
            str bid_order_id, ask_order_id
            bint orders_created = False
        # Number of pair of orders to track for hanging orders
        number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0

        if len(proposal.buys) > 0:
            if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
                price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
                                   f"{buy.price.normalize()} {self.quote_asset}"
                                   for buy in proposal.buys]
                self.logger().info(
                    f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
                    f"at (Size, Price): {price_quote_str}"
                )
            for idx, buy in enumerate(proposal.buys):
                bid_order_id = self.c_buy_with_specific_market(
                    self._market_info,
                    buy.size,
                    order_type=self._limit_order_type,
                    price=buy.price,
                    expiration_seconds=expiration_seconds
                )
                orders_created = True
                if idx < number_of_pairs:
                    order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
                    if order:
                        self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
                            CreatedPairOfOrders(order, None))
        if len(proposal.sells) > 0:
            if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
                price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
                                   f"{sell.price.normalize()} {self.quote_asset}"
                                   for sell in proposal.sells]
                self.logger().info(
                    f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
                    f"orders at (Size, Price): {price_quote_str}"
                )
            for idx, sell in enumerate(proposal.sells):
                ask_order_id = self.c_sell_with_specific_market(
                    self._market_info,
                    sell.size,
                    order_type=self._limit_order_type,
                    price=sell.price,
                    expiration_seconds=expiration_seconds
                )
                orders_created = True
                if idx < number_of_pairs:
                    order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
                    if order:
                        self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
        if orders_created:
            self.c_set_timers()

    def execute_orders_proposal(self, proposal: Proposal):
        self.c_execute_orders_proposal(proposal)

    cdef c_set_timers(self):
        cdef double next_cycle = self._current_timestamp + self.order_refresh_time
        if self._create_timestamp <= self._current_timestamp:
            self._create_timestamp = next_cycle
        if self._cancel_timestamp <= self._current_timestamp:
            self._cancel_timestamp = min(self._create_timestamp, next_cycle)

    def set_timers(self):
        self.c_set_timers()

    def notify_hb_app(self, msg: str):
        if self._hb_app_notification:
            super().notify_hb_app(msg)

    def dump_debug_variables(self):
        market = self._market_info.market
        mid_price = self.get_price()
        spread = Decimal(str(self.c_get_spread()))

        best_ask = mid_price + spread / 2
        new_ask = self._reservation_price + self._optimal_spread / 2
        best_bid = mid_price - spread / 2
        new_bid = self._reservation_price - self._optimal_spread / 2

        vol = self.get_volatility()
        mid_price_variance = vol ** 2

        if not os.path.exists(self._debug_csv_path):
            df_header = pd.DataFrame([('mid_price',
                                       'best_bid',
                                       'best_ask',
                                       'reservation_price',
                                       'optimal_spread',
                                       'optimal_bid',
                                       'optimal_ask',
                                       'optimal_bid_to_mid_%',
                                       'optimal_ask_to_mid_%',
                                       'current_inv',
                                       'target_inv',
                                       'time_left_fraction',
                                       'mid_price std_dev',
                                       'risk_factor',
                                       'gamma',
                                       'alpha',
                                       'kappa',
                                       'eta',
                                       'volatility',
                                       'mid_price_variance',
                                       'inventory_target_pct')])
            df_header.to_csv(self._debug_csv_path, mode='a', header=False, index=False)

        if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:
            time_left_fraction = self._execution_state.time_left / self._execution_state.closing_time
        else:
            time_left_fraction = None

        df = pd.DataFrame([(mid_price,
                            best_bid,
                            best_ask,
                            self._reservation_price,
                            self._optimal_spread,
                            self._optimal_bid,
                            self._optimal_ask,
                            (mid_price - (self._reservation_price - self._optimal_spread / 2)) / mid_price,
                            ((self._reservation_price + self._optimal_spread / 2) - mid_price) / mid_price,
                            market.get_balance(self.base_asset),
                            self.c_calculate_target_inventory(),
                            time_left_fraction,
                            self._avg_vol.current_value,
                            self.gamma,
                            self._alpha,
                            self._kappa,
                            self.eta,
                            vol,
                            mid_price_variance,
                            self.inventory_target_base_pct)])
        df.to_csv(self._debug_csv_path, mode='a', header=False, index=False)



from hummingbot.core.time_iterator cimport TimeIterator


cdef class TransactionTracker(TimeIterator):
    cdef:
        dict _tx_time_limits

    cdef c_start_tx_tracking(self, str tx_id, float timeout_seconds)
    cdef c_stop_tx_tracking(self, str tx_id)
    cdef bint c_is_tx_tracked(self, str tx_id)
    cdef c_did_timeout_tx(self, str tx_id)
    cdef c_process_tx_timeouts(self)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class PrivateAccountV3Api(_message.Message):
    __slots__ = ("vcoinName", "coinId", "balanceAmount", "balanceAmountChange", "frozenAmount", "frozenAmountChange", "type", "time")
    VCOINNAME_FIELD_NUMBER: _ClassVar[int]
    COINID_FIELD_NUMBER: _ClassVar[int]
    BALANCEAMOUNT_FIELD_NUMBER: _ClassVar[int]
    BALANCEAMOUNTCHANGE_FIELD_NUMBER: _ClassVar[int]
    FROZENAMOUNT_FIELD_NUMBER: _ClassVar[int]
    FROZENAMOUNTCHANGE_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    TIME_FIELD_NUMBER: _ClassVar[int]
    vcoinName: str
    coinId: str
    balanceAmount: str
    balanceAmountChange: str
    frozenAmount: str
    frozenAmountChange: str
    type: str
    time: int
    def __init__(self, vcoinName: _Optional[str] = ..., coinId: _Optional[str] = ..., balanceAmount: _Optional[str] = ..., balanceAmountChange: _Optional[str] = ..., frozenAmount: _Optional[str] = ..., frozenAmountChange: _Optional[str] = ..., type: _Optional[str] = ..., time: _Optional[int] = ...) -> None: ...



# distutils: language=c++

from hummingbot.core.clock cimport Clock
from hummingbot.core.pubsub cimport PubSub


cdef class TimeIterator(PubSub):
    cdef:
        double _current_timestamp
        Clock _clock

    cdef c_start(self, Clock clock, double timestamp)
    cdef c_stop(self, Clock clock)
    cdef c_tick(self, double timestamp)



import threading
from typing import TYPE_CHECKING

import pandas as pd

from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.core.data_type.common import PriceType
from hummingbot.core.utils.async_utils import safe_ensure_future

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class TickerCommand:
    def ticker(self,  # type: HummingbotApplication
               live: bool = False,
               exchange: str = None,
               market: str = None):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.ticker, live, exchange, market)
            return
        safe_ensure_future(self.show_ticker(live, exchange, market))

    async def show_ticker(self,  # type: HummingbotApplication
                          live: bool = False,
                          exchange: str = None,
                          market: str = None):
        if len(self.trading_core.markets.keys()) == 0:
            self.notify("\n This command can only be used while a strategy is running")
            return
        if exchange is not None:
            if exchange not in self.trading_core.markets:
                self.notify("\n Please select a valid exchange from the running strategy")
                return
            market_connector = self.trading_core.markets[exchange]
        else:
            market_connector = list(self.trading_core.markets.values())[0]
        if market is not None:
            market = market.upper()
            if market not in market_connector.order_books:
                self.notify("\n Please select a valid trading pair from the running strategy")
                return
            trading_pair, order_book = market, market_connector.order_books[market]
        else:
            trading_pair, order_book = next(iter(market_connector.order_books.items()))

        def get_ticker():
            columns = ["Best Bid", "Best Ask", "Mid Price", "Last Trade"]
            data = [[
                float(market_connector.get_price_by_type(trading_pair, PriceType.BestBid)),
                float(market_connector.get_price_by_type(trading_pair, PriceType.BestAsk)),
                float(market_connector.get_price_by_type(trading_pair, PriceType.MidPrice)),
                float(market_connector.get_price_by_type(trading_pair, PriceType.LastTrade))
            ]]
            ticker_df = pd.DataFrame(data=data, columns=columns)
            ticker_df_str = format_df_for_printout(ticker_df, self.client_config_map.tables_format)
            return f"   Market: {market_connector.name}\n{ticker_df_str}"

        if live:
            await self.stop_live_update()
            self.app.live_updates = True
            while self.app.live_updates:
                await self.cls_display_delay(get_ticker() + "\n\n Press escape key to stop update.", 1)
            self.notify("Stopped live ticker display update.")
        else:
            self.notify(get_ticker())



import hashlib
import random
import time
from decimal import Decimal
from typing import Literal, Optional

import base58
from pydantic import BaseModel, field_validator, model_validator

from hummingbot.client.settings import AllConnectorSettings
from hummingbot.core.data_type.common import TradeType


class ExecutorConfigBase(BaseModel):
    id: str = None  # Make ID optional
    type: Literal["position_executor", "dca_executor", "grid_executor", "order_executor",
                  "xemm_executor", "arbitrage_executor", "twap_executor"]
    timestamp: Optional[float] = None
    controller_id: str = "main"

    @field_validator("timestamp", mode="before")
    @classmethod
    def validate_timestamp(cls, value: Optional[float]) -> float:
        if value is None:
            # Use current time if timestamp is not provided
            return time.time()
        return value

    @model_validator(mode="after")
    def set_id(self):
        if self.id is None:
            # Use timestamp from values if available, else current time
            unique_component = random.randint(0, 99999)
            raw_id = f"{self.timestamp}-{unique_component}"
            hashed_id = hashlib.sha256(raw_id.encode()).digest()  # Get bytes
            self.id = base58.b58encode(hashed_id).decode()  # Base58 encode
        return self


class ConnectorPair(BaseModel):
    connector_name: str
    trading_pair: str

    def is_amm_connector(self) -> bool:
        return self.connector_name in sorted(
            AllConnectorSettings.get_gateway_amm_connector_names()
        )

    class Config:
        frozen = True  # This makes the model immutable and thus hashable

    def __iter__(self):
        yield self.connector_name
        yield self.trading_pair


class PositionSummary(BaseModel):
    connector_name: str
    trading_pair: str
    volume_traded_quote: Decimal
    side: TradeType
    amount: Decimal
    breakeven_price: Decimal
    unrealized_pnl_quote: Decimal
    realized_pnl_quote: Decimal
    cum_fees_quote: Decimal

    @property
    def amount_quote(self) -> Decimal:
        return self.amount * self.breakeven_price

    @property
    def global_pnl_quote(self) -> Decimal:
        return self.unrealized_pnl_quote + self.realized_pnl_quote - self.cum_fees_quote



from hummingbot.connector.exchange_base cimport ExchangeBase
from .asset_price_delegate cimport AssetPriceDelegate

cdef class APIAssetPriceDelegate(AssetPriceDelegate):
    cdef:
        ExchangeBase _market
        object _custom_api_feed



import re
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import TYPE_CHECKING, Dict, List, Literal, Optional, Union

from pydantic import ConfigDict, Field, SecretStr, field_validator
from pyinjective.async_client_v2 import AsyncClient
from pyinjective.composer_v2 import Composer
from pyinjective.core.broadcaster import (
    MessageBasedTransactionFeeCalculator,
    SimulatedTransactionFeeCalculator,
    TransactionFeeCalculator,
)
from pyinjective.core.network import Network
from pyinjective.wallet import PrivateKey

from hummingbot.client.config.config_data_types import BaseClientModel, BaseConnectorConfigMap
from hummingbot.connector.exchange.injective_v2 import injective_constants as CONSTANTS
from hummingbot.connector.exchange.injective_v2.data_sources.injective_grantee_data_source import (
    InjectiveGranteeDataSource,
)
from hummingbot.connector.exchange.injective_v2.data_sources.injective_read_only_data_source import (
    InjectiveReadOnlyDataSource,
)
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

if TYPE_CHECKING:
    from hummingbot.connector.exchange.injective_v2.data_sources.injective_data_source import InjectiveDataSource

CENTRALIZED = False
EXAMPLE_PAIR = "INJ-USDT"

DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0"),
    taker_percent_fee_decimal=Decimal("0"),
)

TESTNET_NODES = ["lb", "sentry"]


class InjectiveFeeCalculatorMode(BaseClientModel, ABC):
    @abstractmethod
    def create_calculator(
        self,
        client: AsyncClient,
        composer: Composer,
        gas_price: Optional[int] = None,
        gas_limit_adjustment_multiplier: Optional[Decimal] = None,
    ) -> Network:
        pass


class InjectiveSimulatedTransactionFeeCalculatorMode(InjectiveFeeCalculatorMode):
    name: Literal["simulated_transaction_fee_calculator"] = "simulated_transaction_fee_calculator"
    model_config = ConfigDict(title="simulated_transaction_fee_calculator")

    def create_calculator(
            self,
            client: AsyncClient,
            composer: Composer,
            gas_price: Optional[int] = None,
            gas_limit_adjustment_multiplier: Optional[Decimal] = None,
    ) -> TransactionFeeCalculator:
        return SimulatedTransactionFeeCalculator(
            client=client,
            composer=composer,
            gas_price=gas_price,
            gas_limit_adjustment_multiplier=gas_limit_adjustment_multiplier,
        )


class InjectiveMessageBasedTransactionFeeCalculatorMode(InjectiveFeeCalculatorMode):
    name: Literal["message_based_transaction_fee_calculator"] = "message_based_transaction_fee_calculator"
    model_config = ConfigDict(title="message_based_transaction_fee_calculator")

    def create_calculator(
            self,
            client: AsyncClient,
            composer: Composer,
            gas_price: Optional[int] = None,
            gas_limit_adjustment_multiplier: Optional[Decimal] = None,
    ) -> TransactionFeeCalculator:
        return MessageBasedTransactionFeeCalculator.new_using_gas_heuristics(
            client=client,
            composer=composer,
            gas_price=gas_price,
        )


FEE_CALCULATOR_MODES = {
    InjectiveSimulatedTransactionFeeCalculatorMode.model_config["title"]: InjectiveSimulatedTransactionFeeCalculatorMode,
    InjectiveMessageBasedTransactionFeeCalculatorMode.model_config["title"]: InjectiveMessageBasedTransactionFeeCalculatorMode,
}


class InjectiveNetworkMode(BaseClientModel, ABC):
    @abstractmethod
    def network(self) -> Network:
        pass


class InjectiveMainnetNetworkMode(InjectiveNetworkMode):
    model_config = ConfigDict(title="mainnet_network")

    def network(self) -> Network:
        return Network.mainnet()

    def rate_limits(self) -> List[RateLimit]:
        return CONSTANTS.PUBLIC_NODE_RATE_LIMITS


class InjectiveTestnetNetworkMode(InjectiveNetworkMode):
    testnet_node: str = Field(
        default="lb",
        json_schema_extra={
            "prompt": f"Enter the testnet node you want to connect to ({'/'.join(TESTNET_NODES)})",
            "prompt_on_new": True}
    )
    model_config = ConfigDict(title="testnet_network")

    @field_validator("testnet_node", mode="before")
    @classmethod
    def validate_node(cls, v: str):
        if v not in TESTNET_NODES:
            raise ValueError(f"{v} is not a valid node ({TESTNET_NODES})")
        return v

    def network(self) -> Network:
        return Network.testnet(node=self.testnet_node)

    def rate_limits(self) -> List[RateLimit]:
        return CONSTANTS.PUBLIC_NODE_RATE_LIMITS


class InjectiveCustomNetworkMode(InjectiveNetworkMode):
    lcd_endpoint: str = Field(
        default=...,
        json_schema_extra={"prompt": "Enter the network lcd_endpoint", "prompt_on_new": True},
    )
    tm_websocket_endpoint: str = Field(
        default=...,
        json_schema_extra={"prompt": "Enter the network tm_websocket_endpoint", "prompt_on_new": True},
    )
    grpc_endpoint: str = Field(
        default=...,
        json_schema_extra={"prompt": "Enter the network grpc_endpoint", "prompt_on_new": True},
    )
    grpc_exchange_endpoint: str = Field(
        default=...,
        json_schema_extra={"prompt": "Enter the network grpc_exchange_endpoint", "prompt_on_new": True},
    )
    grpc_explorer_endpoint: str = Field(
        default=...,
        json_schema_extra={"prompt": "Enter the network grpc_explorer_endpoint", "prompt_on_new": True},
    )
    chain_stream_endpoint: str = Field(
        default=...,
        json_schema_extra={"prompt": "Enter the network chain_stream_endpoint", "prompt_on_new": True},
    )
    chain_id: str = Field(
        default=...,
        json_schema_extra={"prompt": "Enter the network chain_id", "prompt_on_new": True},
    )
    env: str = Field(
        default=...,
        json_schema_extra={"prompt": "Enter the network environment name", "prompt_on_new": True},
    )
    model_config = ConfigDict(title="custom_network")

    def network(self) -> Network:
        return Network.custom(
            lcd_endpoint=self.lcd_endpoint,
            tm_websocket_endpoint=self.tm_websocket_endpoint,
            grpc_endpoint=self.grpc_endpoint,
            grpc_exchange_endpoint=self.grpc_exchange_endpoint,
            grpc_explorer_endpoint=self.grpc_explorer_endpoint,
            chain_stream_endpoint=self.chain_stream_endpoint,
            chain_id=self.chain_id,
            env=self.env,
            official_tokens_list_url=Network.mainnet().official_tokens_list_url,
        )

    def rate_limits(self) -> List[RateLimit]:
        return CONSTANTS.CUSTOM_NODE_RATE_LIMITS


NETWORK_MODES = {
    InjectiveMainnetNetworkMode.model_config["title"]: InjectiveMainnetNetworkMode,
    InjectiveTestnetNetworkMode.model_config["title"]: InjectiveTestnetNetworkMode,
    InjectiveCustomNetworkMode.model_config["title"]: InjectiveCustomNetworkMode,
}

# Captures a 12 or 24-word BIP39 seed phrase
RE_SEED_PHRASE = re.compile(r"^(?:[a-z]+(?: [a-z]+){11}|[a-z]+(?: [a-z]+){23})$")


class InjectiveAccountMode(BaseClientModel, ABC):

    @abstractmethod
    def create_data_source(
            self,
            network: Network,
            rate_limits: List[RateLimit],
            fee_calculator_mode: InjectiveFeeCalculatorMode,
    ) -> "InjectiveDataSource":
        pass


class InjectiveDelegatedAccountMode(InjectiveAccountMode):
    private_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Injective trading account private key or seed phrase",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    subaccount_index: int = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Injective trading account subaccount index",
            "prompt_on_new": True,
        }
    )
    granter_address: str = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter the Injective address of the granter account (portfolio account)",
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    granter_subaccount_index: int = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter the Injective granter subaccount index (portfolio subaccount index)",
            "prompt_on_new": True,
        }
    )

    @field_validator("private_key", mode="before")
    @classmethod
    def validate_network(cls, v: str):
        # Both seed phrase and hex private keys supported
        if isinstance(v, str):
            v = v.strip()
            if RE_SEED_PHRASE.match(v):
                private_key = PrivateKey.from_mnemonic(v)
                return private_key.to_hex()
        return v
    model_config = ConfigDict(title="delegate_account")

    def create_data_source(
            self,
            network: Network,
            rate_limits: List[RateLimit],
            fee_calculator_mode: InjectiveFeeCalculatorMode,
    ) -> "InjectiveDataSource":
        return InjectiveGranteeDataSource(
            private_key=self.private_key.get_secret_value(),
            subaccount_index=self.subaccount_index,
            granter_address=self.granter_address,
            granter_subaccount_index=self.granter_subaccount_index,
            network=network,
            rate_limits=rate_limits,
            fee_calculator_mode=fee_calculator_mode,
        )


class InjectiveReadOnlyAccountMode(InjectiveAccountMode):
    model_config = ConfigDict(title="read_only_account")

    def create_data_source(
            self,
            network: Network,
            rate_limits: List[RateLimit],
            fee_calculator_mode: InjectiveFeeCalculatorMode,
    ) -> "InjectiveDataSource":
        return InjectiveReadOnlyDataSource(
            network=network,
            rate_limits=rate_limits,
        )


ACCOUNT_MODES = {
    InjectiveDelegatedAccountMode.model_config["title"]: InjectiveDelegatedAccountMode,
    InjectiveReadOnlyAccountMode.model_config["title"]: InjectiveReadOnlyAccountMode,
}


class InjectiveConfigMap(BaseConnectorConfigMap):
    # Setting a default dummy configuration to allow the bot to create a dummy instance to fetch all trading pairs
    connector: str = "injective_v2"
    receive_connector_configuration: bool = Field(default=True)
    network: Union[tuple(NETWORK_MODES.values())] = Field(
        default=InjectiveMainnetNetworkMode(),
        json_schema_extra={
            "prompt": f"Select the network ({'/'.join(list(NETWORK_MODES.keys()))})",
            "prompt_on_new": True},
    )
    account_type: Union[tuple(ACCOUNT_MODES.values())] = Field(
        default=InjectiveReadOnlyAccountMode(),
        json_schema_extra={
            "prompt": f"Select the account type ({'/'.join(list(ACCOUNT_MODES.keys()))})",
            "prompt_on_new": True},
    )
    fee_calculator: Union[tuple(FEE_CALCULATOR_MODES.values())] = Field(
        default=InjectiveMessageBasedTransactionFeeCalculatorMode(),
        discriminator="name",
        json_schema_extra={
            "prompt": f"Select the fee calculator ({'/'.join(list(FEE_CALCULATOR_MODES.keys()))})",
            "prompt_on_new": True},
    )
    model_config = ConfigDict(title="injective_v2")

    @field_validator("network", mode="before")
    @classmethod
    def validate_network(cls, v: Union[(str, Dict) + tuple(NETWORK_MODES.values())]):
        if isinstance(v, tuple(NETWORK_MODES.values()) + (Dict,)):
            sub_model = v
        elif v not in NETWORK_MODES:
            raise ValueError(
                f"Invalid network, please choose a value from {list(NETWORK_MODES.keys())}."
            )
        else:
            sub_model = NETWORK_MODES[v].model_construct()
        return sub_model

    @field_validator("account_type", mode="before")
    @classmethod
    def validate_account_type(cls, v: Union[(str, Dict) + tuple(ACCOUNT_MODES.values())]):
        if isinstance(v, tuple(ACCOUNT_MODES.values()) + (Dict,)):
            sub_model = v
        elif v not in ACCOUNT_MODES:
            raise ValueError(
                f"Invalid account type, please choose a value from {list(ACCOUNT_MODES.keys())}."
            )
        else:
            sub_model = ACCOUNT_MODES[v].model_construct()
        return sub_model

    @field_validator("fee_calculator", mode="before")
    @classmethod
    def validate_fee_calculator(cls, v: Union[(str, Dict) + tuple(FEE_CALCULATOR_MODES.values())]):
        if isinstance(v, tuple(FEE_CALCULATOR_MODES.values()) + (Dict,)):
            sub_model = v
        elif v not in FEE_CALCULATOR_MODES:
            raise ValueError(
                f"Invalid fee calculator, please choose a value from {list(FEE_CALCULATOR_MODES.keys())}."
            )
        else:
            sub_model = FEE_CALCULATOR_MODES[v].model_construct()
        return sub_model

    def create_data_source(self):
        return self.account_type.create_data_source(
            network=self.network.network(),
            rate_limits=self.network.rate_limits(),
            fee_calculator_mode=self.fee_calculator,
        )


KEYS = InjectiveConfigMap.model_construct()



#!/usr/bin/env python

import asyncio
from typing import TYPE_CHECKING

from hummingbot.core.utils.async_utils import safe_ensure_future

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class ExitCommand:
    def exit(self,  # type: HummingbotApplication
             force: bool = False):
        safe_ensure_future(self.exit_loop(force), loop=self.ev_loop)

    async def exit_loop(self,  # type: HummingbotApplication
                        force: bool = False):
        if self.trading_core.strategy_task is not None and not self.trading_core.strategy_task.cancelled():
            self.trading_core.strategy_task.cancel()
        if force is False:
            success = await self.trading_core.cancel_outstanding_orders()
            if not success:
                self.notify('Wind down process terminated: Failed to cancel all outstanding orders. '
                            '\nYou may need to manually cancel remaining orders by logging into your chosen exchanges'
                            '\n\nTo force exit the app, enter "exit -f"')
                return
            # Freeze screen 1 second for better UI
            await asyncio.sleep(1)

        if self.trading_core.gateway_monitor is not None:
            self.trading_core.gateway_monitor.stop_monitor()

        self.notify("Winding down notifiers...")
        for notifier in self.trading_core.notifiers:
            notifier.stop()

        self.app.exit()
        self.mqtt_stop()



from decimal import Decimal
from typing import List

import pandas as pd

from hummingbot.core.data_type.common import TradeType
from hummingbot.strategy_v2.backtesting.executor_simulator_base import ExecutorSimulation, ExecutorSimulatorBase
from hummingbot.strategy_v2.executors.dca_executor.data_types import DCAExecutorConfig, DCAMode
from hummingbot.strategy_v2.models.executors import CloseType


class DCAExecutorSimulator(ExecutorSimulatorBase):

    @staticmethod
    def break_even_price_at_index(prices: List[Decimal], amounts: List[Decimal], index: int) -> Decimal:
        total_amount = sum(amounts[:index + 1])
        total_quote = sum([amounts[i] * prices[i] for i in range(index + 1)])
        return total_quote / total_amount

    def simulate(self, df: pd.DataFrame, config: DCAExecutorConfig, trade_cost: float) -> ExecutorSimulation:
        if config.mode == DCAMode.TAKER:
            raise NotImplementedError("Taker mode is not supported in DCAExecutorSimulator")
        potential_dca_stages = []
        side_multiplier = 1 if config.side == TradeType.BUY else -1
        last_timestamp = df['timestamp'].max()
        tl = config.time_limit if config.time_limit else None
        tl_timestamp = config.timestamp + tl if tl else last_timestamp

        # Trailing stop parameters
        trailing_sl_trigger_pct = config.trailing_stop.activation_price if config.trailing_stop else None
        trailing_sl_delta_pct = config.trailing_stop.trailing_delta if config.trailing_stop else None

        # Filter dataframe based on the conditions
        df_filtered = df[:tl_timestamp].copy()
        df_filtered['net_pnl_pct'] = 0.0
        df_filtered['net_pnl_quote'] = 0.0
        df_filtered['cum_fees_quote'] = 0.0
        df_filtered['filled_amount_quote'] = 0.0
        df_filtered['current_position_average_price'] = float(config.prices[0])

        for i in range(len(config.prices)):
            is_last_order = i == len(config.prices) - 1
            price = config.prices[i]
            amount = config.amounts_quote[i]
            break_even_price = DCAExecutorSimulator.break_even_price_at_index(config.prices, config.amounts_quote, i) if i > 0 else price

            entry_condition = (df_filtered['close'] <= price) if config.side == TradeType.BUY else (df_filtered['close'] >= price)
            entry_timestamp = df_filtered[entry_condition]['timestamp'].min()
            if pd.isna(entry_timestamp):
                break
            returns_df = df_filtered[entry_timestamp:]
            returns = returns_df['close'].pct_change().fillna(0)
            cumulative_returns = (((1 + returns).cumprod() - 1) * side_multiplier) - trade_cost
            take_profit_timestamp = None
            stop_loss_timestamp = None
            trailing_sl_timestamp = None
            next_order_timestamp = None

            # Trailing stop logic
            if trailing_sl_trigger_pct is not None and trailing_sl_delta_pct is not None:
                trailing_stop_activation_price = break_even_price * (1 + trailing_sl_trigger_pct * side_multiplier)
                trailing_stop_condition = None
                if config.side == TradeType.BUY:
                    ts_activated_condition = returns_df["close"] >= trailing_stop_activation_price
                    if ts_activated_condition.any():
                        ts_activated_condition = ts_activated_condition.cumsum() > 0
                        with pd.option_context('mode.chained_assignment', None):
                            returns_df.loc[ts_activated_condition, "ts_trigger_price"] = (returns_df[ts_activated_condition]["close"] * float(1 - trailing_sl_delta_pct)).cummax()
                        trailing_stop_condition = returns_df['close'] <= returns_df['ts_trigger_price']
                else:
                    ts_activated_condition = returns_df["close"] <= trailing_stop_activation_price
                    if ts_activated_condition.any():
                        ts_activated_condition = ts_activated_condition.cumsum() > 0
                        with pd.option_context('mode.chained_assignment', None):
                            returns_df.loc[ts_activated_condition, "ts_trigger_price"] = (returns_df[ts_activated_condition]["close"] * float(1 + trailing_sl_delta_pct)).cummin()
                        trailing_stop_condition = returns_df['close'] >= returns_df['ts_trigger_price']
                trailing_sl_timestamp = returns_df[trailing_stop_condition]['timestamp'].min() if trailing_stop_condition is not None else None

            if config.take_profit:
                take_profit_price = break_even_price * (1 + config.take_profit * side_multiplier)
                take_profit_condition = returns_df['close'] >= take_profit_price if config.side == TradeType.BUY else returns_df['close'] <= take_profit_price
                take_profit_timestamp = returns_df[take_profit_condition]['timestamp'].min()

            if is_last_order and config.stop_loss:
                stop_loss_price = break_even_price * (1 - config.stop_loss * side_multiplier)
                stop_loss_condition = returns_df['low'] <= stop_loss_price if config.side == TradeType.BUY else returns_df['high'] >= stop_loss_price
                stop_loss_timestamp = returns_df[stop_loss_condition]['timestamp'].min()
            else:
                next_order_condition = returns_df['close'] <= config.prices[i + 1] if config.side == TradeType.BUY else returns_df['close'] >= config.prices[i + 1]
                next_order_timestamp = returns_df[next_order_condition]['timestamp'].min()

            close_timestamp = min([timestamp for timestamp in [take_profit_timestamp, stop_loss_timestamp,
                                                               trailing_sl_timestamp, last_timestamp, next_order_timestamp] if not pd.isna(timestamp)])

            if close_timestamp == take_profit_timestamp:
                close_type = CloseType.TAKE_PROFIT
            elif close_timestamp == stop_loss_timestamp:
                close_type = CloseType.STOP_LOSS
            elif close_timestamp == trailing_sl_timestamp:
                close_type = CloseType.TRAILING_STOP
            elif close_timestamp == next_order_timestamp:
                close_type = None
            else:
                close_type = CloseType.TIME_LIMIT

            df_filtered[f'filled_amount_quote_{i}'] = 0.0
            df_filtered[f'net_pnl_quote_{i}'] = 0.0
            potential_dca_stages.append({
                'level': i,
                'entry_timestamp': entry_timestamp,
                'price': float(price),
                'amount': float(amount),
                'break_even_price': float(break_even_price),
                'close_timestamp': close_timestamp,
                'close_type': close_type,
                'cumulative_returns': cumulative_returns
            })

        if len(potential_dca_stages) == 0:
            return ExecutorSimulation(config=config, executor_simulation=df_filtered, close_type=CloseType.TIME_LIMIT)

        close_type = None

        for i, dca_stage in enumerate(potential_dca_stages):
            if dca_stage['close_type'] is None:
                df_filtered.loc[entry_timestamp:, f'filled_amount_quote_{i}'] = dca_stage['amount']
                df_filtered.loc[entry_timestamp:, f'net_pnl_quote_{i}'] = dca_stage['cumulative_returns'] * dca_stage['amount']
                df_filtered.loc[entry_timestamp:, 'current_position_average_price'] = dca_stage['break_even_price']
            else:
                df_filtered.loc[entry_timestamp:, f'filled_amount_quote_{i}'] = dca_stage['amount']
                df_filtered.loc[entry_timestamp:, f'net_pnl_quote_{i}'] = dca_stage['cumulative_returns'] * dca_stage['amount']
                df_filtered.loc[entry_timestamp:, 'current_position_average_price'] = dca_stage['break_even_price']
                close_type = dca_stage['close_type']
                last_timestamp = dca_stage['close_timestamp']
                break

        df_filtered = df_filtered[:last_timestamp].copy()
        df_filtered['filled_amount_quote'] = sum([df_filtered[f'filled_amount_quote_{i}'] for i in range(len(potential_dca_stages))])
        df_filtered['net_pnl_quote'] = sum([df_filtered[f'net_pnl_quote_{i}'] for i in range(len(potential_dca_stages))])
        df_filtered['cum_fees_quote'] = trade_cost * df_filtered['filled_amount_quote']
        df_filtered.loc[df_filtered["filled_amount_quote"] > 0, "net_pnl_pct"] = df_filtered["net_pnl_quote"] / df_filtered["filled_amount_quote"]
        df_filtered.loc[df_filtered.index[-1], "filled_amount_quote"] = df_filtered["filled_amount_quote"].iloc[-1] * 2

        if close_type is None:
            close_type = CloseType.FAILED

        # Construct and return ExecutorSimulation object
        simulation = ExecutorSimulation(
            config=config,
            executor_simulation=df_filtered,
            close_type=close_type
        )
        return simulation



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, NoReturn, Optional

from hummingbot.connector.exchange.bitget import bitget_constants as CONSTANTS, bitget_web_utils as web_utils
from hummingbot.connector.exchange.bitget.bitget_auth import BitgetAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest, WSPlainTextRequest, WSResponse
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.bitget.bitget_exchange import BitgetExchange


class BitgetAPIUserStreamDataSource(UserStreamTrackerDataSource):
    """
    Data source for retrieving user stream data from the Bitget exchange via WebSocket APIs.
    """

    _logger: Optional[HummingbotLogger] = None

    def __init__(
        self,
        auth: BitgetAuth,
        trading_pairs: List[str],
        connector: 'BitgetExchange',
        api_factory: WebAssistantsFactory,
    ) -> None:
        super().__init__()
        self._auth = auth
        self._trading_pairs = trading_pairs
        self._connector = connector
        self._api_factory = api_factory
        self._ping_task: Optional[asyncio.Task] = None

    async def _authenticate(self, websocket_assistant: WSAssistant) -> None:
        """
        Authenticates user to websocket
        """
        await websocket_assistant.send(
            WSJSONRequest({
                "op": "login",
                "args": [self._auth.get_ws_auth_payload()]
            })
        )
        response: WSResponse = await websocket_assistant.receive()
        message = response.data

        if (message["event"] != "login" and message["code"] != "0"):
            self.logger().error(
                f"Error authenticating the private websocket connection. Response message {message}"
            )
            raise IOError("Private websocket connection authentication failed")

    async def _parse_pong_message(self) -> None:
        self.logger().debug("PING-PONG message for user stream completed")

    async def _process_message_for_unknown_channel(
        self,
        event_message: Dict[str, Any]
    ) -> None:
        if event_message == CONSTANTS.PUBLIC_WS_PONG_RESPONSE:
            await self._parse_pong_message()
        elif "event" in event_message:
            if event_message["event"] == "error":
                message = event_message.get("msg", "Unknown error")
                error_code = event_message.get("code", "Unknown code")
                self.logger().error(
                    f"Failed to subscribe to private channels: {message} ({error_code})"
                )

            if event_message["event"] == "subscribe":
                channel: str = event_message["arg"]["channel"]
                self.logger().info(f"Subscribed to private channel: {channel.upper()}")
        else:
            self.logger().warning(f"Message for unknown channel received: {event_message}")

    async def _process_event_message(
        self,
        event_message: Dict[str, Any],
        queue: asyncio.Queue
    ) -> None:
        if "arg" in event_message and "action" in event_message:
            queue.put_nowait(event_message)
        else:
            await self._process_message_for_unknown_channel(event_message)

    async def _subscribe_channels(self, websocket_assistant: WSAssistant) -> None:
        try:
            subscription_topics = []

            for channel in [CONSTANTS.WS_ACCOUNT_ENDPOINT, CONSTANTS.WS_FILL_ENDPOINT]:
                subscription_topics.append({
                    "instType": "SPOT",
                    "channel": channel,
                    "coin": "default"
                })

            for trading_pair in self._trading_pairs:
                subscription_topics.append({
                    "instType": "SPOT",
                    "channel": CONSTANTS.WS_ORDERS_ENDPOINT,
                    "instId": await self._connector.exchange_symbol_associated_to_pair(trading_pair)
                })
            await websocket_assistant.send(
                WSJSONRequest({
                    "op": "subscribe",
                    "args": subscription_topics
                })
            )
            self.logger().info("Subscribed to private channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to private channels...")
            raise

    async def _connected_websocket_assistant(self) -> WSAssistant:
        websocket_assistant: WSAssistant = await self._api_factory.get_ws_assistant()

        await websocket_assistant.connect(
            ws_url=web_utils.private_ws_url(),
            message_timeout=CONSTANTS.SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE
        )
        await self._authenticate(websocket_assistant)

        return websocket_assistant

    async def _send_ping(self, websocket_assistant: WSAssistant) -> None:
        await websocket_assistant.send(
            WSPlainTextRequest(CONSTANTS.PUBLIC_WS_PING_REQUEST)
        )

    async def send_interval_ping(self, websocket_assistant: WSAssistant) -> None:
        """
        Coroutine to send PING messages periodically.

        :param websocket_assistant: The websocket assistant to use to send the PING message.
        """
        try:
            while True:
                await self._send_ping(websocket_assistant)
                await asyncio.sleep(CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        except asyncio.CancelledError:
            self.logger().info("Interval PING task cancelled")
            raise
        except Exception:
            self.logger().exception("Error sending interval PING")

    async def listen_for_user_stream(self, output: asyncio.Queue) -> NoReturn:
        while True:
            try:
                self._ws_assistant = await self._connected_websocket_assistant()
                await self._subscribe_channels(websocket_assistant=self._ws_assistant)
                self._ping_task = asyncio.create_task(self.send_interval_ping(self._ws_assistant))
                await self._process_websocket_messages(
                    websocket_assistant=self._ws_assistant,
                    queue=output
                )
            except asyncio.CancelledError:
                raise
            except ConnectionError as connection_exception:
                self.logger().warning(
                    f"The websocket connection was closed ({connection_exception})"
                )
            except Exception:
                self.logger().exception(
                    "Unexpected error while listening to user stream. Retrying after 5 seconds..."
                )
                await self._sleep(1.0)
            finally:
                if self._ping_task is not None:
                    self._ping_task.cancel()
                    try:
                        await self._ping_task
                    except asyncio.CancelledError:
                        pass
                    self._ping_task = None
                await self._on_user_stream_interruption(websocket_assistant=self._ws_assistant)
                self._ws_assistant = None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: trade.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    descriptor_pool as _descriptor_pool,
    symbol_database as _symbol_database,
)
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0btrade.proto\x12\x05trade"J\n\x0b\x43redentials\x12\x15\n\raccess_key_id\x18\x01 \x01(\t\x12\x11\n\tsignature\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x04"\xcb\x01\n\x0cOrderRequest\x12\x1e\n\x03new\x18\x01 \x01(\x0b\x32\x0f.trade.NewOrderH\x00\x12$\n\x06\x63\x61ncel\x18\x02 \x01(\x0b\x32\x12.trade.CancelOrderH\x00\x12$\n\x06modify\x18\x03 \x01(\x0b\x32\x12.trade.ModifyOrderH\x00\x12%\n\theartbeat\x18\x04 \x01(\x0b\x32\x10.trade.HeartbeatH\x00\x12\x1f\n\x02mc\x18\x05 \x01(\x0b\x32\x11.trade.MassCancelH\x00\x42\x07\n\x05inner"\x99\x03\n\x08NewOrder\x12\x17\n\x0f\x63lient_order_id\x18\x01 \x01(\x04\x12\x12\n\nrequest_id\x18\x02 \x01(\x04\x12\x11\n\tmarket_id\x18\x03 \x01(\x04\x12\x12\n\x05price\x18\x04 \x01(\x04H\x00\x88\x01\x01\x12\x10\n\x08quantity\x18\x05 \x01(\x04\x12\x19\n\x04side\x18\x06 \x01(\x0e\x32\x0b.trade.Side\x12)\n\rtime_in_force\x18\x07 \x01(\x0e\x32\x12.trade.TimeInForce\x12$\n\norder_type\x18\x08 \x01(\x0e\x32\x10.trade.OrderType\x12\x15\n\rsubaccount_id\x18\t \x01(\x04\x12>\n\x15self_trade_prevention\x18\n \x01(\x0e\x32\x1a.trade.SelfTradePreventionH\x01\x88\x01\x01\x12"\n\tpost_only\x18\x0b \x01(\x0e\x32\x0f.trade.PostOnly\x12\x1c\n\x14\x63\x61ncel_on_disconnect\x18\x0c \x01(\x08\x42\x08\n\x06_priceB\x18\n\x16_self_trade_prevention"d\n\x0b\x43\x61ncelOrder\x12\x11\n\tmarket_id\x18\x01 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x15\n\rsubaccount_id\x18\x04 \x01(\x04"\x8b\x02\n\x0bModifyOrder\x12\x11\n\tmarket_id\x18\x01 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x11\n\tnew_price\x18\x04 \x01(\x04\x12\x14\n\x0cnew_quantity\x18\x05 \x01(\x04\x12\x15\n\rsubaccount_id\x18\x06 \x01(\x04\x12>\n\x15self_trade_prevention\x18\x07 \x01(\x0e\x32\x1a.trade.SelfTradePreventionH\x00\x88\x01\x01\x12"\n\tpost_only\x18\x08 \x01(\x0e\x32\x0f.trade.PostOnlyB\x18\n\x16_self_trade_prevention"\x86\x01\n\nMassCancel\x12\x15\n\rsubaccount_id\x18\x01 \x01(\x04\x12\x12\n\nrequest_id\x18\x02 \x01(\x04\x12\x16\n\tmarket_id\x18\x03 \x01(\x04H\x00\x88\x01\x01\x12\x1e\n\x04side\x18\x04 \x01(\x0e\x32\x0b.trade.SideH\x01\x88\x01\x01\x42\x0c\n\n_market_idB\x07\n\x05_side"2\n\tHeartbeat\x12\x12\n\nrequest_id\x18\x01 \x01(\x04\x12\x11\n\ttimestamp\x18\x02 \x01(\x04"\xcb\x03\n\rOrderResponse\x12%\n\x07new_ack\x18\x01 \x01(\x0b\x32\x12.trade.NewOrderAckH\x00\x12+\n\ncancel_ack\x18\x02 \x01(\x0b\x32\x15.trade.CancelOrderAckH\x00\x12+\n\nmodify_ack\x18\x03 \x01(\x0b\x32\x15.trade.ModifyOrderAckH\x00\x12+\n\nnew_reject\x18\x04 \x01(\x0b\x32\x15.trade.NewOrderRejectH\x00\x12\x31\n\rcancel_reject\x18\x05 \x01(\x0b\x32\x18.trade.CancelOrderRejectH\x00\x12\x31\n\rmodify_reject\x18\x06 \x01(\x0b\x32\x18.trade.ModifyOrderRejectH\x00\x12\x1b\n\x04\x66ill\x18\x07 \x01(\x0b\x32\x0b.trade.FillH\x00\x12%\n\theartbeat\x18\x08 \x01(\x0b\x32\x10.trade.HeartbeatH\x00\x12(\n\x08position\x18\t \x01(\x0b\x32\x14.trade.AssetPositionH\x00\x12/\n\x0fmass_cancel_ack\x18\n \x01(\x0b\x32\x14.trade.MassCancelAckH\x00\x42\x07\n\x05inner"\xe5\x02\n\x0bNewOrderAck\x12\x13\n\x0bmsg_seq_num\x18\x01 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x19\n\x11\x65xchange_order_id\x18\x04 \x01(\x04\x12\x11\n\tmarket_id\x18\x05 \x01(\x04\x12\x12\n\x05price\x18\x06 \x01(\x04H\x00\x88\x01\x01\x12\x10\n\x08quantity\x18\x07 \x01(\x04\x12\x19\n\x04side\x18\x08 \x01(\x0e\x32\x0b.trade.Side\x12)\n\rtime_in_force\x18\t \x01(\x0e\x32\x12.trade.TimeInForce\x12$\n\norder_type\x18\n \x01(\x0e\x32\x10.trade.OrderType\x12\x15\n\rtransact_time\x18\x0b \x01(\x04\x12\x15\n\rsubaccount_id\x18\x0c \x01(\x04\x12\x1c\n\x14\x63\x61ncel_on_disconnect\x18\r \x01(\x08\x42\x08\n\x06_price"\xeb\x02\n\x0e\x43\x61ncelOrderAck\x12\x13\n\x0bmsg_seq_num\x18\x01 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x15\n\rtransact_time\x18\x04 \x01(\x04\x12\x15\n\rsubaccount_id\x18\x05 \x01(\x04\x12,\n\x06reason\x18\x06 \x01(\x0e\x32\x1c.trade.CancelOrderAck.Reason\x12\x11\n\tmarket_id\x18\x07 \x01(\x04\x12\x19\n\x11\x65xchange_order_id\x18\x08 \x01(\x04"\x8c\x01\n\x06Reason\x12\x10\n\x0cUNCLASSIFIED\x10\x00\x12\x0e\n\nDISCONNECT\x10\x01\x12\r\n\tREQUESTED\x10\x02\x12\x07\n\x03IOC\x10\x03\x12\x0f\n\x0bSTP_RESTING\x10\x04\x12\x12\n\x0eSTP_AGGRESSING\x10\x05\x12\x0f\n\x0bMASS_CANCEL\x10\x06\x12\x12\n\x0ePOSITION_LIMIT\x10\x07"\x88\x02\n\x0eModifyOrderAck\x12\x13\n\x0bmsg_seq_num\x18\x01 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x15\n\rtransact_time\x18\x04 \x01(\x04\x12\x1a\n\x12remaining_quantity\x18\x05 \x01(\x04\x12\x15\n\rsubaccount_id\x18\x06 \x01(\x04\x12\x11\n\tmarket_id\x18\x07 \x01(\x04\x12\r\n\x05price\x18\x08 \x01(\x04\x12\x10\n\x08quantity\x18\t \x01(\x04\x12\x1b\n\x13\x63umulative_quantity\x18\n \x01(\x04\x12\x19\n\x11\x65xchange_order_id\x18\x0b \x01(\x04"\x87\x02\n\rMassCancelAck\x12\x13\n\x0bmsg_seq_num\x18\x01 \x01(\x04\x12\x15\n\rsubaccount_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x15\n\rtransact_time\x18\x04 \x01(\x04\x12\x30\n\x06reason\x18\x06 \x01(\x0e\x32\x1b.trade.MassCancelAck.ReasonH\x00\x88\x01\x01\x12\x1d\n\x15total_affected_orders\x18\x07 \x01(\r"C\n\x06Reason\x12\x10\n\x0cUNCLASSIFIED\x10\x00\x12\x15\n\x11INVALID_MARKET_ID\x10\x01\x12\x10\n\x0cINVALID_SIDE\x10\x02\x42\t\n\x07_reason"\xb1\x07\n\x0eNewOrderReject\x12\x13\n\x0bmsg_seq_num\x18\x01 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x15\n\rtransact_time\x18\x04 \x01(\x04\x12\x15\n\rsubaccount_id\x18\x05 \x01(\x04\x12,\n\x06reason\x18\x06 \x01(\x0e\x32\x1c.trade.NewOrderReject.Reason\x12\x11\n\tmarket_id\x18\x07 \x01(\x04\x12\x12\n\x05price\x18\x08 \x01(\x04H\x00\x88\x01\x01\x12\x10\n\x08quantity\x18\t \x01(\x04\x12\x19\n\x04side\x18\n \x01(\x0e\x32\x0b.trade.Side\x12)\n\rtime_in_force\x18\x0b \x01(\x0e\x32\x12.trade.TimeInForce\x12$\n\norder_type\x18\x0c \x01(\x0e\x32\x10.trade.OrderType"\xd1\x04\n\x06Reason\x12\x10\n\x0cUNCLASSIFIED\x10\x00\x12\x14\n\x10INVALID_QUANTITY\x10\x01\x12\x15\n\x11INVALID_MARKET_ID\x10\x02\x12\x16\n\x12\x44UPLICATE_ORDER_ID\x10\x03\x12\x10\n\x0cINVALID_SIDE\x10\x04\x12\x19\n\x15INVALID_TIME_IN_FORCE\x10\x05\x12\x16\n\x12INVALID_ORDER_TYPE\x10\x06\x12\x15\n\x11INVALID_POST_ONLY\x10\x07\x12!\n\x1dINVALID_SELF_TRADE_PREVENTION\x10\x08\x12\x12\n\x0eUNKNOWN_TRADER\x10\t\x12!\n\x1dPRICE_WITH_MARKET_LIMIT_ORDER\x10\n\x12\x1f\n\x1bPOST_ONLY_WITH_MARKET_ORDER\x10\x0b\x12\x1e\n\x1aPOST_ONLY_WITH_INVALID_TIF\x10\x0c\x12\x1a\n\x16\x45XCEEDED_SPOT_POSITION\x10\r\x12\x1d\n\x19NO_OPPOSING_RESTING_ORDER\x10\x0e\x12\x19\n\x15POST_ONLY_WOULD_TRADE\x10\x0f\x12\x16\n\x12\x44ID_NOT_FULLY_FILL\x10\x10\x12\x1e\n\x1aONLY_ORDER_CANCEL_ACCEPTED\x10\x11\x12$\n PROTECTION_PRICE_WOULD_NOT_TRADE\x10\x12\x12\x16\n\x12NO_REFERENCE_PRICE\x10\x13\x12\x15\n\x11SLIPPAGE_TOO_HIGH\x10\x14\x12\x16\n\x12OUTSIDE_PRICE_BAND\x10\x15\x42\x08\n\x06_price"\x8f\x02\n\x11\x43\x61ncelOrderReject\x12\x13\n\x0bmsg_seq_num\x18\x01 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x15\n\rtransact_time\x18\x04 \x01(\x04\x12\x15\n\rsubaccount_id\x18\x05 \x01(\x04\x12/\n\x06reason\x18\x06 \x01(\x0e\x32\x1f.trade.CancelOrderReject.Reason\x12\x11\n\tmarket_id\x18\x07 \x01(\x04"F\n\x06Reason\x12\x10\n\x0cUNCLASSIFIED\x10\x00\x12\x15\n\x11INVALID_MARKET_ID\x10\x01\x12\x13\n\x0fORDER_NOT_FOUND\x10\x02"\xf4\x03\n\x11ModifyOrderReject\x12\x13\n\x0bmsg_seq_num\x18\x01 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x02 \x01(\x04\x12\x12\n\nrequest_id\x18\x03 \x01(\x04\x12\x15\n\rtransact_time\x18\x04 \x01(\x04\x12\x15\n\rsubaccount_id\x18\x05 \x01(\x04\x12/\n\x06reason\x18\x06 \x01(\x0e\x32\x1f.trade.ModifyOrderReject.Reason\x12\x11\n\tmarket_id\x18\x07 \x01(\x04"\xaa\x02\n\x06Reason\x12\x10\n\x0cUNCLASSIFIED\x10\x00\x12\x14\n\x10INVALID_QUANTITY\x10\x01\x12\x15\n\x11INVALID_MARKET_ID\x10\x02\x12\x13\n\x0fORDER_NOT_FOUND\x10\x03\x12\x0f\n\x0bINVALID_IFM\x10\x04\x12\x15\n\x11INVALID_POST_ONLY\x10\x05\x12!\n\x1dINVALID_SELF_TRADE_PREVENTION\x10\x06\x12\x12\n\x0eUNKNOWN_TRADER\x10\x07\x12\x1a\n\x16\x45XCEEDED_SPOT_POSITION\x10\x08\x12\x19\n\x15POST_ONLY_WOULD_TRADE\x10\t\x12\x1e\n\x1aONLY_ORDER_CANCEL_ACCEPTED\x10\x11\x12\x16\n\x12OUTSIDE_PRICE_BAND\x10\x0b"\xe8\x02\n\x04\x46ill\x12\x13\n\x0bmsg_seq_num\x18\x01 \x01(\x04\x12\x11\n\tmarket_id\x18\x02 \x01(\x04\x12\x17\n\x0f\x63lient_order_id\x18\x03 \x01(\x04\x12\x19\n\x11\x65xchange_order_id\x18\x04 \x01(\x04\x12\x12\n\nfill_price\x18\x05 \x01(\x04\x12\x15\n\rfill_quantity\x18\x06 \x01(\x04\x12\x17\n\x0fleaves_quantity\x18\x07 \x01(\x04\x12\x15\n\rtransact_time\x18\x08 \x01(\x04\x12\x15\n\rsubaccount_id\x18\t \x01(\x04\x12\x1b\n\x13\x63umulative_quantity\x18\n \x01(\x04\x12\x19\n\x04side\x18\x0b \x01(\x0e\x32\x0b.trade.Side\x12\x1b\n\x13\x61ggressor_indicator\x18\x0c \x01(\x08\x12+\n\tfee_ratio\x18\r \x01(\x0b\x32\x18.trade.FixedPointDecimal\x12\x10\n\x08trade_id\x18\x0e \x01(\x04"7\n\x11\x46ixedPointDecimal\x12\x10\n\x08mantissa\x18\x01 \x01(\x03\x12\x10\n\x08\x65xponent\x18\x02 \x01(\x05"|\n\rAssetPosition\x12\x15\n\rsubaccount_id\x18\x01 \x01(\x04\x12\x10\n\x08\x61sset_id\x18\x02 \x01(\x04\x12\x1e\n\x05total\x18\x03 \x01(\x0b\x32\x0f.trade.RawUnits\x12"\n\tavailable\x18\x04 \x01(\x0b\x32\x0f.trade.RawUnits"F\n\x08RawUnits\x12\r\n\x05word0\x18\x01 \x01(\x04\x12\r\n\x05word1\x18\x02 \x01(\x04\x12\r\n\x05word2\x18\x03 \x01(\x04\x12\r\n\x05word3\x18\x04 \x01(\x04"\x85\x01\n\tBootstrap\x12\x1b\n\x04\x64one\x18\x01 \x01(\x0b\x32\x0b.trade.DoneH\x00\x12\'\n\x07resting\x18\x02 \x01(\x0b\x32\x14.trade.RestingOrdersH\x00\x12)\n\x08position\x18\x03 \x01(\x0b\x32\x15.trade.AssetPositionsH\x00\x42\x07\n\x05inner"4\n\rRestingOrders\x12#\n\x06orders\x18\x01 \x03(\x0b\x32\x13.trade.RestingOrder"9\n\x0e\x41ssetPositions\x12\'\n\tpositions\x18\x01 \x03(\x0b\x32\x14.trade.AssetPosition"7\n\x04\x44one\x12\x1c\n\x14latest_transact_time\x18\x01 \x01(\x04\x12\x11\n\tread_only\x18\x02 \x01(\x08"\xe9\x02\n\x0cRestingOrder\x12\x17\n\x0f\x63lient_order_id\x18\x01 \x01(\x04\x12\x19\n\x11\x65xchange_order_id\x18\x02 \x01(\x04\x12\x11\n\tmarket_id\x18\x03 \x01(\x04\x12\r\n\x05price\x18\x04 \x01(\x04\x12\x16\n\x0eorder_quantity\x18\x05 \x01(\x04\x12\x19\n\x04side\x18\x06 \x01(\x0e\x32\x0b.trade.Side\x12)\n\rtime_in_force\x18\x07 \x01(\x0e\x32\x12.trade.TimeInForce\x12$\n\norder_type\x18\x08 \x01(\x0e\x32\x10.trade.OrderType\x12\x1a\n\x12remaining_quantity\x18\t \x01(\x04\x12\x11\n\trest_time\x18\n \x01(\x04\x12\x15\n\rsubaccount_id\x18\x0b \x01(\x04\x12\x1b\n\x13\x63umulative_quantity\x18\x0c \x01(\x04\x12\x1c\n\x14\x63\x61ncel_on_disconnect\x18\r \x01(\x08*\x18\n\x04Side\x12\x07\n\x03\x42ID\x10\x00\x12\x07\n\x03\x41SK\x10\x01*N\n\x0bTimeInForce\x12\x17\n\x13IMMEDIATE_OR_CANCEL\x10\x00\x12\x14\n\x10GOOD_FOR_SESSION\x10\x01\x12\x10\n\x0c\x46ILL_OR_KILL\x10\x02*D\n\tOrderType\x12\t\n\x05LIMIT\x10\x00\x12\x10\n\x0cMARKET_LIMIT\x10\x01\x12\x1a\n\x16MARKET_WITH_PROTECTION\x10\x02*V\n\x13SelfTradePrevention\x12\x12\n\x0e\x43\x41NCEL_RESTING\x10\x00\x12\x15\n\x11\x43\x41NCEL_AGGRESSING\x10\x01\x12\x14\n\x10\x41LLOW_SELF_TRADE\x10\x02*%\n\x08PostOnly\x12\x0c\n\x08\x44ISABLED\x10\x00\x12\x0b\n\x07\x45NABLED\x10\x01\x42\x12Z\x03go/\xaa\x02\nCube.Tradeb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "trade_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS is False:
    _globals["DESCRIPTOR"]._options = None
    _globals["DESCRIPTOR"]._serialized_options = b"Z\003go/\252\002\nCube.Trade"
    _globals["_SIDE"]._serialized_start = 6011
    _globals["_SIDE"]._serialized_end = 6035
    _globals["_TIMEINFORCE"]._serialized_start = 6037
    _globals["_TIMEINFORCE"]._serialized_end = 6115
    _globals["_ORDERTYPE"]._serialized_start = 6117
    _globals["_ORDERTYPE"]._serialized_end = 6185
    _globals["_SELFTRADEPREVENTION"]._serialized_start = 6187
    _globals["_SELFTRADEPREVENTION"]._serialized_end = 6273
    _globals["_POSTONLY"]._serialized_start = 6275
    _globals["_POSTONLY"]._serialized_end = 6312
    _globals["_CREDENTIALS"]._serialized_start = 22
    _globals["_CREDENTIALS"]._serialized_end = 96
    _globals["_ORDERREQUEST"]._serialized_start = 99
    _globals["_ORDERREQUEST"]._serialized_end = 302
    _globals["_NEWORDER"]._serialized_start = 305
    _globals["_NEWORDER"]._serialized_end = 714
    _globals["_CANCELORDER"]._serialized_start = 716
    _globals["_CANCELORDER"]._serialized_end = 816
    _globals["_MODIFYORDER"]._serialized_start = 819
    _globals["_MODIFYORDER"]._serialized_end = 1086
    _globals["_MASSCANCEL"]._serialized_start = 1089
    _globals["_MASSCANCEL"]._serialized_end = 1223
    _globals["_HEARTBEAT"]._serialized_start = 1225
    _globals["_HEARTBEAT"]._serialized_end = 1275
    _globals["_ORDERRESPONSE"]._serialized_start = 1278
    _globals["_ORDERRESPONSE"]._serialized_end = 1737
    _globals["_NEWORDERACK"]._serialized_start = 1740
    _globals["_NEWORDERACK"]._serialized_end = 2097
    _globals["_CANCELORDERACK"]._serialized_start = 2100
    _globals["_CANCELORDERACK"]._serialized_end = 2463
    _globals["_CANCELORDERACK_REASON"]._serialized_start = 2323
    _globals["_CANCELORDERACK_REASON"]._serialized_end = 2463
    _globals["_MODIFYORDERACK"]._serialized_start = 2466
    _globals["_MODIFYORDERACK"]._serialized_end = 2730
    _globals["_MASSCANCELACK"]._serialized_start = 2733
    _globals["_MASSCANCELACK"]._serialized_end = 2996
    _globals["_MASSCANCELACK_REASON"]._serialized_start = 2918
    _globals["_MASSCANCELACK_REASON"]._serialized_end = 2985
    _globals["_NEWORDERREJECT"]._serialized_start = 2999
    _globals["_NEWORDERREJECT"]._serialized_end = 3944
    _globals["_NEWORDERREJECT_REASON"]._serialized_start = 3341
    _globals["_NEWORDERREJECT_REASON"]._serialized_end = 3934
    _globals["_CANCELORDERREJECT"]._serialized_start = 3947
    _globals["_CANCELORDERREJECT"]._serialized_end = 4218
    _globals["_CANCELORDERREJECT_REASON"]._serialized_start = 4148
    _globals["_CANCELORDERREJECT_REASON"]._serialized_end = 4218
    _globals["_MODIFYORDERREJECT"]._serialized_start = 4221
    _globals["_MODIFYORDERREJECT"]._serialized_end = 4721
    _globals["_MODIFYORDERREJECT_REASON"]._serialized_start = 4423
    _globals["_MODIFYORDERREJECT_REASON"]._serialized_end = 4721
    _globals["_FILL"]._serialized_start = 4724
    _globals["_FILL"]._serialized_end = 5084
    _globals["_FIXEDPOINTDECIMAL"]._serialized_start = 5086
    _globals["_FIXEDPOINTDECIMAL"]._serialized_end = 5141
    _globals["_ASSETPOSITION"]._serialized_start = 5143
    _globals["_ASSETPOSITION"]._serialized_end = 5267
    _globals["_RAWUNITS"]._serialized_start = 5269
    _globals["_RAWUNITS"]._serialized_end = 5339
    _globals["_BOOTSTRAP"]._serialized_start = 5342
    _globals["_BOOTSTRAP"]._serialized_end = 5475
    _globals["_RESTINGORDERS"]._serialized_start = 5477
    _globals["_RESTINGORDERS"]._serialized_end = 5529
    _globals["_ASSETPOSITIONS"]._serialized_start = 5531
    _globals["_ASSETPOSITIONS"]._serialized_end = 5588
    _globals["_DONE"]._serialized_start = 5590
    _globals["_DONE"]._serialized_end = 5645
    _globals["_RESTINGORDER"]._serialized_start = 5648
    _globals["_RESTINGORDER"]._serialized_end = 6009
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from decimal import Decimal
from typing import List, Literal, Optional

from pydantic import BaseModel, ConfigDict

from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.strategy_v2.executors.data_types import ExecutorConfigBase


class TrailingStop(BaseModel):
    activation_price: Decimal
    trailing_delta: Decimal


class TripleBarrierConfig(BaseModel):
    stop_loss: Optional[Decimal] = None
    take_profit: Optional[Decimal] = None
    time_limit: Optional[int] = None
    trailing_stop: Optional[TrailingStop] = None
    open_order_type: OrderType = OrderType.LIMIT
    take_profit_order_type: OrderType = OrderType.MARKET
    stop_loss_order_type: OrderType = OrderType.MARKET
    time_limit_order_type: OrderType = OrderType.MARKET
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def new_instance_with_adjusted_volatility(self, volatility_factor: float) -> TripleBarrierConfig:
        new_trailing_stop = None
        if self.trailing_stop is not None:
            new_trailing_stop = TrailingStop(
                activation_price=self.trailing_stop.activation_price * Decimal(volatility_factor),
                trailing_delta=self.trailing_stop.trailing_delta * Decimal(volatility_factor)
            )

        return TripleBarrierConfig(
            stop_loss=self.stop_loss * Decimal(volatility_factor) if self.stop_loss is not None else None,
            take_profit=self.take_profit * Decimal(volatility_factor) if self.take_profit is not None else None,
            time_limit=self.time_limit,
            trailing_stop=new_trailing_stop,
            open_order_type=self.open_order_type,
            take_profit_order_type=self.take_profit_order_type,
            stop_loss_order_type=self.stop_loss_order_type,
            time_limit_order_type=self.time_limit_order_type
        )


class PositionExecutorConfig(ExecutorConfigBase):
    type: Literal["position_executor"] = "position_executor"
    trading_pair: str
    connector_name: str
    side: TradeType
    entry_price: Optional[Decimal] = None
    amount: Decimal
    triple_barrier_config: TripleBarrierConfig = TripleBarrierConfig()
    leverage: int = 1
    activation_bounds: Optional[List[Decimal]] = None
    level_id: Optional[str] = None
    model_config = ConfigDict(arbitrary_types_allowed=True)



# distutils: language=c++
# distutils: sources=hummingbot/core/cpp/OrderBookEntry.cpp

from typing import Iterator

from cython.operator cimport address as ref, dereference as deref, postincrement as inc
from hummingbot.core.data_type.OrderBookEntry cimport OrderBookEntry
from libcpp.set cimport set
from libcpp.vector cimport vector

from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book_row import OrderBookRow

cdef class CompositeOrderBook(OrderBook):
    """
    Record orders that are bought during back testing and used to simulate order book consumption without modifying
    the actual order book.
    Override the order book bid_entries, ask_entries methods to return the composite order book entries
    """
    def __init__(self, order_book: OrderBook = None):
        super().__init__()
        self._traded_order_book = OrderBook()

    @property
    def traded_order_book(self) -> OrderBook:
        return self._traded_order_book

    def clear_traded_order_book(self):
        self._traded_order_book._bid_book.clear()
        self._traded_order_book._ask_book.clear()

    def record_filled_order(self, order_fill_event):
        cdef:
            vector[OrderBookEntry] cpp_bids
            vector[OrderBookEntry] cpp_asks
            set[OrderBookEntry].reverse_iterator bid_order_it = self._traded_order_book._bid_book.rbegin()
            set[OrderBookEntry].iterator ask_order_it = self._traded_order_book._ask_book.begin()
            OrderBookEntry entry

        price = order_fill_event.price
        amount = float(order_fill_event.amount)
        timestamp = order_fill_event.timestamp

        if order_fill_event.trade_type is TradeType.BUY:
            while ask_order_it != self._traded_order_book._ask_book.end():
                entry = deref(ask_order_it)
                # price is in the order book, sum the amount
                if entry.getPrice() == price:
                    amount += entry.getAmount()
                    break
                # price is outside of the ask price range, break and insert the new filled ask order into the ask book
                elif entry.getPrice() > price:
                    break
                # price is further up the ask price range, continue searching
                elif entry.getPrice() < price:
                    inc(ask_order_it)
            cpp_asks.push_back(OrderBookEntry(price, amount, timestamp))

        elif order_fill_event.trade_type is TradeType.SELL:
            while bid_order_it != self._traded_order_book._bid_book.rend():
                entry = deref(bid_order_it)
                if entry.getPrice() == price:
                    amount += entry.getAmount()
                    break
                # price is further down the bid price range, continue searching
                elif entry.getPrice() > price:
                    inc(bid_order_it)
                # price is outside of the bid price range, break and insert the new filled order into the bid book
                elif entry.getPrice() < price:
                    break
            cpp_bids.push_back(OrderBookEntry(price, amount, timestamp))

        self._traded_order_book.c_apply_diffs(cpp_bids, cpp_asks, timestamp)

    def original_bid_entries(self) -> Iterator[OrderBookRow]:
        return super().bid_entries()

    def original_ask_entries(self) -> Iterator[OrderBookRow]:
        return super().ask_entries()

    def bid_entries(self) -> Iterator[OrderBookRow]:
        cdef:
            set[OrderBookEntry].reverse_iterator order_it = self._bid_book.rbegin()
            set[OrderBookEntry].reverse_iterator traded_order_it = self._traded_order_book._bid_book.rbegin()
            OrderBookEntry traded_order_entry
            OrderBookEntry original_order_entry
            vector[OrderBookEntry] cpp_asks_changes
            vector[OrderBookEntry] cpp_bids_changes

        while order_it != self._bid_book.rend():
            original_order_entry = deref(order_it)
            original_order_price = original_order_entry.getPrice()
            original_order_amount = original_order_entry.getAmount()
            original_order_update_id = original_order_entry.getUpdateId()

            while traded_order_it != self._traded_order_book._bid_book.rend():
                traded_order_entry = deref(traded_order_it)
                traded_order_price = traded_order_entry.getPrice()
                traded_order_amount = traded_order_entry.getAmount()
                traded_order_update_id = traded_order_entry.getUpdateId()

                # Found matching price for the recorded filled order, return composite order book row
                if traded_order_price == original_order_price:
                    composite_amount = original_order_amount - traded_order_amount
                    if composite_amount > 0:
                        yield OrderBookRow(original_order_price, composite_amount, original_order_update_id)
                    else:
                        cpp_bids_changes.push_back(OrderBookEntry(original_order_price,
                                                                  min(original_order_amount, traded_order_amount),
                                                                  traded_order_update_id))
                    inc(traded_order_it)
                    # continue to next original order book row
                    break
                # Recorded filled order price is outside of the bid price range
                elif traded_order_price > original_order_price:
                    # Remove the recorded entry and increment the pointer
                    cpp_bids_changes.push_back(OrderBookEntry(traded_order_price, 0, traded_order_update_id))
                    inc(traded_order_it)
                # Recorded filled order price is within lower end of the bid price range, yield original bid entry
                elif traded_order_price < original_order_price:
                    yield OrderBookRow(original_order_price, original_order_amount, original_order_update_id)
                    break
            else:
                yield OrderBookRow(original_order_price, original_order_amount, original_order_update_id)

            inc(order_it)

        self._traded_order_book.c_apply_diffs(cpp_bids_changes, cpp_asks_changes, self._last_diff_uid)

    def ask_entries(self) -> Iterator[OrderBookRow]:
        cdef:
            set[OrderBookEntry].iterator order_it = self._ask_book.begin()
            set[OrderBookEntry].iterator traded_order_it = self._traded_order_book._ask_book.begin()
            OrderBookEntry original_order_entry
            OrderBookEntry traded_order_entry
            vector[OrderBookEntry] cpp_asks_changes
            vector[OrderBookEntry] cpp_bids_changes

        while order_it != self._ask_book.end():
            original_order_entry = deref(order_it)
            original_order_price = original_order_entry.getPrice()
            original_order_amount = original_order_entry.getAmount()
            original_order_update_id = original_order_entry.getUpdateId()

            while traded_order_it != self._traded_order_book._ask_book.end():
                traded_order_entry = deref(traded_order_it)
                traded_order_price = traded_order_entry.getPrice()
                traded_order_amount = traded_order_entry.getAmount()
                traded_order_update_id = traded_order_entry.getUpdateId()

                if traded_order_price == original_order_price:
                    composite_amount = original_order_amount - traded_order_amount
                    if composite_amount > 0:
                        yield OrderBookRow(original_order_price, composite_amount, original_order_update_id)
                    else:
                        cpp_asks_changes.push_back(OrderBookEntry(original_order_price,
                                                                  min(original_order_amount, traded_order_amount),
                                                                  traded_order_update_id))
                    inc(traded_order_it)
                    # continue to next original order book row
                    break
                # Recorded filled order price is within upper end of the ask price range, yield original ask entry
                elif traded_order_price > original_order_price:
                    yield OrderBookRow(original_order_price, original_order_amount, original_order_update_id)
                    break
                # Recorded filled order price is outside of the ask price range, remove the recorded ask order
                elif traded_order_price < original_order_price:
                    cpp_asks_changes.push_back(OrderBookEntry(traded_order_price, 0, traded_order_update_id))
                    inc(traded_order_it)

            else:
                yield OrderBookRow(original_order_price, original_order_amount, original_order_update_id)

            inc(order_it)

        self._traded_order_book.c_apply_diffs(cpp_bids_changes, cpp_asks_changes, self._last_diff_uid)

    cdef double c_get_price(self, bint is_buy) except? -1:
        cdef:
            set[OrderBookEntry] *book = ref(self._ask_book) if is_buy else ref(self._bid_book)
        if deref(book).size() < 1:
            raise EnvironmentError("Order book is empty - no price quote is possible.")

        ask_it = self.ask_entries()
        bid_it = self.bid_entries()
        try:
            if is_buy:
                best_ask = next(ask_it)
                return best_ask.price
            else:
                best_bid = next(bid_it)
                return best_bid.price
        except Exception:
            raise



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.mexc.com"
HEALTH_CHECK_ENDPOINT = "/api/v3/ping"
CANDLES_ENDPOINT = "/api/v3/klines"

WSS_URL = "wss://wbs-api.mexc.com/ws"

KLINE_ENDPOINT_NAME = "spot@public.kline.v3.api.pb"

INTERVALS = bidict({
    "1m": "1m",
    "5m": "5m",
    "15m": "15m",
    "30m": "30m",
    "1h": "60m",
    "4h": "4h",
    "1d": "1d",
    "1w": "1W",
    "1M": "1M"
})

WS_INTERVALS = {
    "1m": "Min1",
    "5m": "Min5",
    "15m": "Min15",
    "30m": "Min30",
    "1h": "Min60",
    "4h": "Hour4",
    "8h": "Hour8",
    "1d": "Day1",
    "1w": "Week1",
    "1M": "Month1"
}

MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1000

RATE_LIMITS = [
    RateLimit(CANDLES_ENDPOINT, limit=20000, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)]),
    RateLimit(HEALTH_CHECK_ENDPOINT, limit=20000, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)])]



from decimal import Decimal
from typing import Optional

from sqlalchemy import (
    Column,
    Integer,
    Numeric,
    String,
    UniqueConstraint,
)
from sqlalchemy.orm import Session

from hummingbot.model import HummingbotBase


class InventoryCost(HummingbotBase):
    __tablename__ = "InventoryCost"
    __table_args__ = (
        UniqueConstraint("base_asset", "quote_asset"),
    )

    id = Column(Integer, primary_key=True, nullable=False)
    base_asset = Column(String(45), nullable=False)
    quote_asset = Column(String(45), nullable=False)
    base_volume = Column(Numeric(48, 18), nullable=False)
    quote_volume = Column(Numeric(48, 18), nullable=False)

    @classmethod
    def get_record(
        cls, sql_session: Session, base_asset: str, quote_asset: str
    ) -> Optional["InventoryCost"]:
        return (
            sql_session.query(cls)
            .filter(cls.base_asset == base_asset, cls.quote_asset == quote_asset)
            .first()
        )

    @classmethod
    def add_volume(
        cls,
        sql_session: Session,
        base_asset: str,
        quote_asset: str,
        base_volume: Decimal,
        quote_volume: Decimal,
        overwrite: bool = False,
    ) -> None:
        if overwrite:
            update = {
                "base_volume": base_volume,
                "quote_volume": quote_volume,
            }
        else:
            update = {
                "base_volume": cls.base_volume + base_volume,
                "quote_volume": cls.quote_volume + quote_volume,
            }

        rows_updated: int = sql_session.query(cls).filter(
            cls.base_asset == base_asset, cls.quote_asset == quote_asset
        ).update(update)

        if not rows_updated:
            record = InventoryCost(
                base_asset=base_asset,
                quote_asset=quote_asset,
                base_volume=float(base_volume),
                quote_volume=float(quote_volume),
            )
            sql_session.add(record)



import os.path
from typing import List, Tuple

import pandas as pd

from hummingbot import data_path
from hummingbot.client.hummingbot_application import HummingbotApplication
from hummingbot.strategy.avellaneda_market_making import AvellanedaMarketMakingStrategy
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple


async def start(self):
    try:
        c_map = self.strategy_config_map
        exchange = c_map.exchange
        raw_trading_pair = c_map.market

        trading_pair: str = raw_trading_pair
        base, quote = trading_pair.split("-")
        maker_assets: Tuple[str, str] = (base, quote)
        market_names: List[Tuple[str, List[str]]] = [(exchange, [trading_pair])]
        await self.initialize_markets(market_names)
        maker_data = [self.markets[exchange], trading_pair] + list(maker_assets)
        self.market_trading_pair_tuples = [MarketTradingPairTuple(*maker_data)]

        strategy_logging_options = AvellanedaMarketMakingStrategy.OPTION_LOG_ALL

        debug_csv_path = os.path.join(data_path(),
                                      HummingbotApplication.main_application().strategy_file_name.rsplit('.', 1)[0] +
                                      f"_{pd.Timestamp.now().strftime('%Y-%m-%d_%H-%M-%S')}.csv")

        self.strategy = AvellanedaMarketMakingStrategy()
        self.strategy.init_params(
            config_map=c_map,
            market_info=MarketTradingPairTuple(*maker_data),
            logging_options=strategy_logging_options,
            hb_app_notification=True,
            debug_csv_path=debug_csv_path,
            is_debug=False
        )
    except Exception as e:
        self.notify(str(e))
        self.logger().error("Unknown error during initialization.", exc_info=True)



import asyncio
from decimal import Decimal
from typing import List, Optional, Set, Tuple

import pandas as pd
import psutil
import tabulate

from hummingbot.client.config.config_data_types import ClientConfigEnum
from hummingbot.client.performance import PerformanceMetrics
from hummingbot.model.trade_fill import TradeFill

s_decimal_0 = Decimal("0")


def format_bytes(size):
    for unit in ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB"]:
        if abs(size) < 1024.0:
            return f"{size:.2f} {unit}"
        size /= 1024.0
    return f"{size:.2f} YB"


async def start_timer(timer):
    count = 1
    while True:
        count += 1

        mins, sec = divmod(count, 60)
        hour, mins = divmod(mins, 60)
        days, hour = divmod(hour, 24)

        timer.log(f"Uptime: {days:>3} day(s), {hour:02}:{mins:02}:{sec:02}")
        await _sleep(1)


async def _sleep(delay):
    """
    A wrapper function that facilitates patching the sleep in unit tests without affecting the asyncio module
    """
    await asyncio.sleep(delay)


async def start_process_monitor(process_monitor):
    hb_process = psutil.Process()
    while True:
        with hb_process.oneshot():
            threads = hb_process.num_threads()
            process_monitor.log("CPU: {:>5}%, ".format(hb_process.cpu_percent()) +
                                "Mem: {:>10} ({}), ".format(
                                    format_bytes(hb_process.memory_info().vms / threads),
                                    format_bytes(hb_process.memory_info().rss)) +
                                "Threads: {:>3}, ".format(threads)
                                )
        await _sleep(1)


async def start_trade_monitor(trade_monitor):
    from hummingbot.client.hummingbot_application import HummingbotApplication
    hb = HummingbotApplication.main_application()
    trade_monitor.log("Trades: 0, Total P&L: 0.00, Return %: 0.00%")

    while True:
        try:
            if hb.trading_core._strategy_running and hb.trading_core.strategy is not None:
                if all(market.ready for market in hb.trading_core.markets.values()):
                    with hb.trading_core.trade_fill_db.get_new_session() as session:
                        trades: List[TradeFill] = hb._get_trades_from_session(
                            int(hb.init_time * 1e3),
                            session=session,
                            config_file_path=hb.strategy_file_name)
                        if len(trades) > 0:
                            return_pcts = []
                            pnls = []
                            market_info: Set[Tuple[str, str]] = set((t.market, t.symbol) for t in trades)
                            for market, symbol in market_info:
                                cur_trades = [t for t in trades if t.market == market and t.symbol == symbol]
                                cur_balances = await hb.trading_core.get_current_balances(market)
                                perf = await PerformanceMetrics.create(symbol, cur_trades, cur_balances)
                                return_pcts.append(perf.return_pct)
                                pnls.append(perf.total_pnl)
                            avg_return = sum(return_pcts) / len(return_pcts) if len(return_pcts) > 0 else s_decimal_0
                            quote_assets = set(t.symbol.split("-")[1] for t in trades)
                            if len(quote_assets) == 1:
                                total_pnls = f"{PerformanceMetrics.smart_round(sum(pnls))} {list(quote_assets)[0]}"
                            else:
                                total_pnls = "N/A"
                            trade_monitor.log(f"Trades: {len(trades)}, Total P&L: {total_pnls}, "
                                              f"Return %: {avg_return:.2%}")
            await _sleep(2.0)  # sleeping for longer to manage resources
        except asyncio.CancelledError:
            raise
        except Exception:
            hb.logger().exception("start_trade_monitor failed.")
            await _sleep(2.0)


def format_df_for_printout(
    df: pd.DataFrame, table_format: ClientConfigEnum, max_col_width: Optional[int] = None, index: bool = False
) -> str:
    if max_col_width is not None:  # in anticipation of the next release of tabulate which will include maxcolwidth
        max_col_width = max(max_col_width, 4)
        df = df.astype(str).apply(
            lambda s: s.apply(
                lambda e: e if len(e) < max_col_width else f"{e[:max_col_width - 3]}..."
            )
        )
        df.columns = [c if len(c) < max_col_width else f"{c[:max_col_width - 3]}..." for c in df.columns]

    original_preserve_whitespace = tabulate.PRESERVE_WHITESPACE
    tabulate.PRESERVE_WHITESPACE = True
    try:
        formatted_df = tabulate.tabulate(df, tablefmt=table_format, showindex=index, headers="keys")
    finally:
        tabulate.PRESERVE_WHITESPACE = original_preserve_whitespace
    return formatted_df
